# Functions for processing N2 adsorption isotherms
#
import os
import datetime
import numpy as np
from scipy import stats
from collections import namedtuple

gas2liq = 0.001545


def findBETrange(iso):
    """  Extract from isotherm the data in BET range (0.04 < Pr < 0.30).

    :param iso: Curve (2 vectors) containing the adsorption isotherm.
    :return: Return the iso range required and a error message.
    """
    errmsg = ''
    # find best range
    (nvec, npt) = np.shape(iso)
    istart = 0
    iend = npt-1
    for i in range (npt):
        if iso[0][i] > 0.03:
            istart = i
            break
    for i in range (istart+1, npt):
        # stop when Pr > 0.30
        if iso[0][i] > 0.30:
        # stop when Va(1-Pr) is no more increasing with Pr
        # if iso[1][i]*(1-iso[0][i]) < iso[1][i-1]*(1-iso[0][i-1]):
            iend = i
            break
    n = iend - istart + 1
    if n < 7:
        errmsg = 'Not enough data points in BET range (0.04 < Pr < 0.30)'
    else:
        lst = np.hsplit(iso, np.array([istart, iend+1]))
        BET = lst[1]
        return (BET, errmsg)



def calcBET(BET):
    """ Compute a linear regression on the BET transform.

    :param BET: 3D array which contains relative pressure [0], adsorbed
                volume [1] and BET transform [2].
    :return: The results of the BET computation (r_val is the correlation
             coefficient).
    """
    slope, intercept, r_val, p_val, std_err = stats.linregress(BET[0],BET[2])
    C = slope / intercept + 1
    Vm = 1 / (intercept * C)
    SBET = 4.3525 * Vm
    return (slope, intercept, r_val, std_err, C, Vm, SBET)


def processBET(iso, logfilnam=None):
    """ Return the optimum BET values.

        Find, in the relative pressure range 0.04 - 0.30, a window of at least
        seven data points, giving the best correlation coefficient.

    :param iso: Curve (2 vectors) containing the adsorption isotherm.
    :param logfilnam: name of the file where the results will be saved.
    :return: errmsg (an error message), bestprm (the BET parameters giving the
             best correlation coefficient), Pri (initial relative pressure),
             Prf (final relative pressure).
    """
    reslist = []
    bestprm = None
    Pri = Prf = None
    BETprop = namedtuple('BETprop', 'w i slope intercept cc std_err C Vm SBET')
    (BET, errmsg) = findBETrange(iso)
    if not errmsg:
        (nvar, npt) = np.shape(BET)
        # Append a row of zeros:
        BET = np.vstack((BET, np.zeros((1, BET.shape[1]))))
        # Calculate BET transform
        for i in range(npt):
            BET[2][i] = 1/(BET[1][i]*(1/BET[0][i]-1))
        bestprm = None
        minpt = 7   # number minimum of points needed to compute SBET
        for w in range(minpt, npt+1):
            for i in range(npt-w+1):
                lst = np.hsplit(BET, np.array([i, i+w+1]))
                RBET = lst[1]
                (slope, intercept, cc, std_err, C, Vm, SBET) = calcBET(RBET)
                res = [w, i, slope, intercept, cc, std_err, C, Vm, SBET]
                reslist.append(res)
                # store the best result in bestprm
                if intercept > 0.0:
                    if bestprm is None:
                        bestprm = BETprop(w, i, slope, intercept, cc, std_err, C, Vm, SBET)
                    else:
                        if cc > bestprm.cc:
                            bestprm = BETprop(w, i, slope, intercept, cc, std_err, C, Vm, SBET)
        if bestprm is not None:
            if logfilnam is not None:
                # save results in log file
                if os.path.exists(logfilnam):
                    fo = open(logfilnam, 'a')
                else:
                    fo = open(logfilnam, 'w')
                # write current date and time from now variable
                headtxt = "#\n#\n#%s\n" % datetime.datetime.now()
                headtxt += "#Results of computation of BET parameters\n"
                headtxt += '#w\t i\t slope\t intercept\t r_val\t std_err\t C\t Vm\t SBET\n'
                fo.write(headtxt)
                for lin in reslist:
                    strlst = []
                    strlst.append("{0}".format(lin[0]))
                    strlst.append("{0}".format(lin[1]))
                    for n in lin[2:]:
                        strlst.append("{0:E}".format(n))
                    strlin = '\t'.join(strlst)
                    fo.write("{0}\n".format(strlin))
                fo.close()
            Pri = BET[0][bestprm.i]
            Prf = BET[0][bestprm.i + bestprm.w - 1]
        else:
            errmsg = 'Unable to calculate BET parameters'
    return (errmsg, bestprm, Pri, Prf)


def BET(iso, ads, logfilnam=None):
    """ Compute BET from an N2 adsorption isotherm (P/Po, Va)

    :param iso: Curve (2 vectors) containing the adsorption isotherm.
    :param ads: String indicating the adsorbent. Currently only nitrogen
                (ads='N2') is implemented.
    :param logfilnam: name of the file where the results will be saved.
    :return: a tuple (err, msg) where err = True if an error occurred,
             and msg contains either the error message or the results.
    """
    errmsg = None
    err = False
    if ads != 'N2':
        err = True
        errmsg = "Only nitrogen adsorption is implemented"
        return err, errmsg

    if iso[0].min() < 0.0 or iso[0].max() > 1.0:
        errmsg = 'Relative pressure out of range'
    if errmsg is None:
        if iso[1].min() < 0.0 or iso[1].max() > 5000:
            errmsg = 'Adsorbed volume out of range'
    if errmsg is not None:
        return err, errmsg

    (msg, bestprm, Pri, Prf) = processBET(iso, logfilnam)
    if msg:
        err = True
    else:
        msg =  'BET Surface Area Report \n\n'
        msg += 'Best fitting obtained for :  {} data points\n'.format(bestprm.w)
        msg += '  First relative pressure : {0:6.3f} \n'.format(Pri)
        msg += '   Last relative pressure : {0:6.3f} \n\n'.format(Prf)
        msg += '         BET Surface Area : {0:7.2f} m2/g\n'.format(bestprm.SBET)
        msg += '                        C : {0:5.1f} \n'.format(bestprm.C)
        msg += '                    Slope :  {0:5.4f} g/cm3\n'.format(bestprm.slope)
        msg += '              Y-Intercept :  {0:4.5f} g/cm3\n'.format(bestprm.intercept)
        msg += '                       Vm :  {0:6.3f} cm3/g\n'.format(bestprm.Vm)
        msg += '  Correlation Coefficient :  {0:4.5f} \n\n'.format(bestprm.cc)
        if iso[0].max() > 0.98:
            msg += '  Adsorbed volume at Psat : {0:7.2f} mL/g\n'.format(iso[1].max())
            msg += '              Pore volume : {0:6.3f} mL/g \n\n'.format(iso[1].max() * gas2liq)
    if logfilnam is not None:
        fo = open(logfilnam, 'a')
        fo.write("\n{0}".format(msg))
        fo.close()
        msg += 'These results are saved in {0}'.format(os.path.basename(logfilnam))
    return err, msg


#- PSD --------------------------------------------------------------

def KelvinRadius(Pr, F):
    """ Return the Kelvin radius according to the relative pressure Pr.

        The returned value is valid only for N2 at 77K.
        r = 2 * gamma * Vm /(RT log(Pr)
        gamma = liquid/vapor surface tension = 8.94 mN/m
        Vm = molar volume of liquid N2 = 34.84 mL/mol
        R = universal gas constant = 8.314 J.K−1.mol−1
        T = 77.3 K

    :param Pr: Relative pressure
    :param F: This is the fraction of pores open at both end (F=0 for desorption).
    :return: The Kelvin radius in A.
    """
    C = -9.62
    return C / ((1+F) * np.log(Pr))



def tFit(Pr):
    """ Return the thickness of adsorbed layer (in A) according to 'Pr'.

        The thickness of adsorbed layer is calculated with 4 different
        equations according to the Pr range in order to be as close a
        possible to experimental data recorded on non porous adsorbents.
    :param Pr: The relative pressure
    :return: The thickness of adsorbed layer (in A)
    """
    t = 0.0
    if Pr < 0.13:
        t = -78.0 * Pr * Pr + 25.0 * Pr + 2.23
    elif Pr >= 0.13 and Pr < 0.55:
        t = 7.23 * Pr + 3.28
    elif Pr >= 0.55 and Pr < 0.7:
        t = 13.15 * Pr * Pr - 7.13 * Pr + 7.24
    elif Pr >= 0.7:
        t = 57.68 * Pr * Pr - 73.19 * Pr + 31.63
    return t

def tHalsey(Pr):
    """ Return the thickness of adsorbed layer (in A) according to 'Pr'.

    :param Pr: The relative pressure
    :return: The thickness of adsorbed layer (in A)
    """
    return 3.54 * (-5.0/np.log(Pr))**0.33333

def tHarkins(Pr):
    """ Return the thickness of adsorbed layer (in A) according to 'Pr'.

    :param Pr: The relative pressure
    :return: The thickness of adsorbed layer (in A)
    """
    return (1.399/(0.034 - np.log10(Pr)))**0.5


def PSDcalc(iso, savename, tcurv='tfit', geom='cylindric'):
    """ Compute the PSD from N2 adsorption isotherm.

        Pore Size Distribution (PSD) is calculated with BJH method.
        The PSD is saved to the file "PSD.txt". This file contains 4 vectors:
        the pore width (Wp), the pore volume (Vp), the pore volume divided
        by the pore width difference (dV/dW) and the cumulative pore volume (Vcum).

    :param iso: Curve (2 vectors) containing the isotherm.
    :param tcurv: String indicating the thickness curve to use. Possible values
                  are "tfit", "halsey" or "harkins".
    :param geom: Pore geometry. Currently only cylindrical pore is implemented.
    :return: A tuple (err, msg) where err = True if an error occurred,
             and msg contains either the error message or information.
    """
    errmsg = None
    err = False
    Pr = iso[0]
    Va = iso[1]
    npt = Pr.size
    if Pr[0] > Pr[npt-1]:
        branch = 'desorption'
        F = 0
    else:
        branch = "adsorption"
        # Reverse data order
        Pr = Pr[::-1]
        Va = Va[::-1]
        F = 1
    imin = 0
    while Pr[imin] > 0.995:
        imin += 1
    imax = npt-1
    while Pr[imax] < 0.3:
        imax -= 1
    Pr = Pr[imin:imax]
    Va = Va[imin:imax]

    npt = Pr.size
    t = np.zeros(npt)
    rp = np.zeros(npt)
    rpm = np.zeros(npt)
    rk = np.zeros(npt)
    Vp = np.zeros(npt)
    Ap = np.zeros(npt)
    SAp = np.zeros(npt)
    SVp = np.zeros(npt)

    for i in range(0, npt):
        rk[i] = KelvinRadius(Pr[i], F)
        if tcurv == 'halsey':
            t[i] = tHalsey(Pr[i])
        elif tcurv == 'harkins':
            t[i] = tHarkins(Pr[i])
        else:
            t[i] = tFit(Pr[i])
        rp[i] = rk[i] + t[i]

    for i in range(1, npt):
        rpm[i] = (rp[i-1] + rp[i]) / 2.0
        dt = t[i-1] - t[i]
        tm = (t[i - 1] + t[i])/2
        dVliq = (Va[i-1] - Va[i]) * gas2liq
        R = (rp[i]/(rk[i]+dt))**2
        Vp[i] = R * (dVliq - dt * SAp[i-1] * 1e-4)
        if Vp[i] < 0.0:
            Vp[i] = 0.0
            Ap[i] = 0.0
            SAp[i] = SAp[i-1]
            SVp[i] = SVp[i-1]
        else:
            Ap[i] = 20000.0 * Vp[i] / rpm[i]
            c = (rpm[i]-tm)/rpm[i]
            SAp[i] = SAp[i-1] + c * Ap[i]
            SVp[i] = SVp[i-1] + Vp[i]

    if branch == 'desorption':
        # Revert the order to have increasing diameters
        rp = np.array(rp[::-1])
        Vp = np.array(Vp[::-1])
        Ap = np.array(Ap[::-1])
        
    # Convert Pore with in nm
    Wp = rp * 0.2
    # Calculation of Dp/dw
    dW = np.diff(Wp)
    Vp = Vp[:-1]
    Vcum = np.cumsum(Vp)
    dVdW = Vp / dW
    Wp = Wp[:-1]

    # Save the Pore Size Distribution in file
    header = "PSD for {} isotherm\n".format(branch)
    header += "Thickness curve = {}\n".format(tcurv)
    header += "plot Wp,dVdW\n"
    header += "plot Wp,Vcum,2\n"
    header += "labX Pore width (nm)\n"
    header += "labY1 dV/dW (mL/g)\n"
    header += "labY2 Cumulative pore volume (mL/g)\n"
    header += "Wp\tVp\tdVdW\tVcum"
    numform = ['%5.1f', '%7.5f', '%7.5f', '%7.4f']
    data = np.stack((Wp, Vp, dVdW, Vcum), axis=-1)
    np.savetxt(savename, data, delimiter='\t', fmt=numform, header=header)
    return err, errmsg



import sys
from PyQt5 import QtWidgets
from pyDataVis.pyDataVis import MainWindow


def run():
    app = QtWidgets.QApplication(sys.argv)
    app.setApplicationName("pyDataVis")
    app.setOrganizationName("Pierre Alphonse")
    main = MainWindow()
    main.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    run()




""" This module contains the script class

    A script is a command.
    There is 4 types of commands:
        
    type 0 = no argument
        BET, BETKr, BETAr, PSD, IRabs, IRtrans

    type 1 = commands with one argument (generally the vector name)
        del X, stats Y, area V1, line V1, lineq V1, revert Y,
        sort Y, delMulX Y, onset Y, fft Y, symbol Y, nosymbol Y

    type 2 = commands with two arguments
        clipup Y1 10
        clipdn Y2 0
        clipx < 10
        clipx > 100
        shift V1 -0.5
        shrink V1 5
        linefit V1 -1
        despike V1 winsiz
        ndec Y 5

    type 3 = commands for operation on vectors or between vectors
        Y2 = Y1
        Y1 = Y1 * 5.0
        Y1 = Y1 / 10
        Y1 = Y1 - Y2
        Y2 = log(Y1)
        Y3 = grad(Y1)
        Y3 = sum(Y1)

    type 4 = rename command
        name Y1 as Y2

    type 5 = new command
        new X 0,10,1
"""

import os
import numpy as np
from scipy import signal

from pyDataVis.utils import isNumber, calcArea, shrinkRows, sortArr, delDupliX, rms, pp
from pyDataVis.convert import convertToAbs, convertToTrans
from pyDataVis.convIso import BET, PSDcalc
from pyDataVis.plotWindow import vectInfo, lineSeg


class script(object):
    """ A script is a command.

        Empty lines and lines starting with # are ignored for execution
        script must be called by a plotWin.
    """

    def __init__(self, parent):
        self.parent = parent
        self.operators = set( '+-*/()' )
        self.npfunc = ['arccos', 'arccosh', 'arcsin', 'arcsinh', 'arctan',
                       'arctanh', 'cos', 'cosh', 'cumsum', 'cumprod', 'exp',
                       'fabs', 'gradient', 'log', 'log10', 'power',
                       'sin', 'sinh', 'sqrt', 'square', 'tan', 'tanh']
        self.typ0Lst = [ 'BET', 'BETKr', 'BETAr', 'IRabs', 'IRtrans',
                         'invertx', 'inverty']
        self.typ1Lst = [ 'delv', 'delb', 'stats', 'area', 'line', 'lineq',
                         'revertb', 'revertv', 'sort', 'deldupx', 'onset',
                         'fft' ]
        self.typ2Lst = [ 'swapv', 'mergeb', 'clipup', 'clipdn', 'clipx',
                         'linefit', 'shift', 'shrink', 'despike', 'ndec',
                         'PSD' ]
        self.typ4Lst = [ 'name' ]
        self.typ5Lst = [ 'newv' ]



    def checkType(self, cmd):
        """ Check the type of the command 'cmd'

        :param cmd: a string containing the command
        :return: a tuple containing the command type and the number of
                 arguments. Return (None,0) if the command was not found.
        """
        cmd = cmd.strip()
        if len(cmd) == 0 or cmd[0] == '#':
            # Empty command
            return (None, 0)

        items = cmd.split(' ')
        nitems = len(items)
        narg = nitems - 1
        # Is it a new command (type = 5)
        if cmd.startswith('newv'):
            return (5, narg)
        # Is it a rename command (type = 4) ?
        if cmd.startswith('name'):
            return (4, narg)
        # Is it an expression or operation command (type = 3) ?
        if cmd.find('=') != -1:
            return (3, narg)
        if nitems == 3:
            # It is a two arg command (type = 2)
            return (2, narg)
        if nitems == 2:
            # It is a one arg command (type = 1)
            return (1, narg)
        if nitems == 1:
           # It is a command without argument
           return (0, narg)
        return (None, -1)


    def lookForCmd(self, cmd):
        """ Check if 'cmd' contains a valid command.

        :param cmd: a string containing the command
        :return: a tuple containing the command type and the expected number
                 number of arguments. Return (None,0) if the command was not
                 found.
        """
        cmd = cmd.strip()
        items = cmd.split(' ')
        if len(items) == 1:
            cmdnam = cmd
        else:
            cmdnam = items[0]
        if cmdnam in self.typ0Lst:
            return (0,0)
        if cmdnam in self.typ1Lst:
            return (1,1)
        if cmdnam in self.typ2Lst:
            return (2,2)
        if cmdnam in self.typ4Lst:
            return (4,3)
        if cmdnam in self.typ5Lst:
            return (5,2)
        return (None,0)



    def dispatch(self, cmd):
        """ Check the type of the command 'cmd'

            Call the corresponding exe function.
        :param cmd: a string containing the command
        :return: a tuple (err, msg).
            The behavior of the calling function when returning is expected to be:
            if err = 0 and mgs = "" update the plot
            if err = 0 and msg != "" shows the message in information window
            if err > 0 and msg != "" shows the message in QMessageBox.warning
            if err < 0 and msg != "" shows the message in information window and update the plot
        """
        if cmd.find('=') != -1:
            # this is an expression
            typ = 3
            narg = -1
        else:
            # If this is not an expression check the command name
            # Remove space after comma
            cmd = cmd.replace(", ", ",")
            typ, narg = self.lookForCmd(cmd)
            if typ is None:
                return (1, "Unknown command")

        typ2, narg2 = self.checkType(cmd)
        if typ2 is None:
            if narg2 == 0:
                return (1, "Empty command")
            else:
                return (1, "Unable to process command")
        if typ == 3:  # expression or operation command
            return self.exeExpression(cmd)

        if narg2 < narg:
            return (1, "Not enough parameters")
        if narg2 > narg:
            return (1, "Too much parameters")
        if typ == 5:  # new command
            return newV(self.parent, cmd)
        elif typ == 4:  # rename command
            return name(self.parent, cmd)
        elif typ == 2:  # command with two argument
            return self.exeTwoArg(cmd)
        elif typ == 0:  # command with no argument
            return self.exeNoArg(cmd)
        else:
            # It is a one arg command
            items = cmd.split(' ')
            # look for the command name
            try:
                cmdidx = self.typ1Lst.index(items[0])
            except ValueError:
                    return (1, "Check the parameters")
            # look for the vector name
            if self.parent.getVinfo(items[1]) is None:
                # vector not found, is it a number ?
                if isNumber(items[1]):
                    if items[0] == "delb":
                        blkno = items[1]
                        nblk =  len(self.parent.blklst)
                        if int(blkno) < 0 and int(blkno) > nblk:
                            return (1, "Wrong block number")
                else:
                    if items[0] == "delb":
                        return (1, "Wrong block number")
                    else:
                        return (1, "Unknown vector name")
            return self.exeOneArg(cmdidx, items[1])


    def exeExpression(self, cmd):
        """ Process the commands containing an expression.

            Expression means that the command contains '='.
        :param cmd: a string containing the command
        :return: a tuple (errno, errmsg).
        """
        items = cmd.split('=')
        vnam = items[0].strip()
        expression = items[1].strip()

        # Build vector name list
        # Build vector ndarray list
        vectLst = []
        vectNam = []
        for i in range(len(self.parent.vectInfolst)):
            for j in range(len(self.parent.vectInfolst[i])):
                vectNam.append(self.parent.vectInfolst[i][j].name)
                vectLst.append(self.parent.blklst[i][j])

        # Check that all the vectors in expression belong to the same data block
        first = True
        for vn in vectNam:
            if vn in expression:
                vinfo = self.parent.getVinfo(vn)
                blkno = vinfo.blkpos
                if first:
                    first = False
                else:
                    if vinfo.blkpos != blkno:
                        return (1, "All vectors must belong "
                                   "to the same datablock")
                    break

        # Check expression for unauthorized functions or unknown vector name
        # The goal is to secure the use of eval. Besides check that all
        # the vectors belong to the same data block.
        blkno = -1
        s = expression
        for c in self.operators:
            s = s.replace(c, ',')
        items = s.split(',')
        for item in items:
            item = item.strip()
            if item:
                if not item in vectNam:
                    if not item in self.npfunc:
                        if not isNumber(item):
                            return (1, "Unknown vector or function")
                else:
                    if blkno == -1:
                        blkno = self.parent.getVinfo(item).blkpos
                    else:
                        if self.parent.getVinfo(item).blkpos != blkno:
                            return (1, "All vectors must belong to the same data block")

        # Prefix, in expression, the function name with 'np.'
        for fn in self.npfunc:
            if fn in expression:
                expression = expression.replace(fn, "np.{0}".format(fn))

        # Build the nam: ndarray dictionary for eval
        dictVec = {vectNam[i]: vectLst[i] for i in range(len(vectNam))}

        try:
            R = eval(expression, globals(), dictVec)
        except SyntaxError:
            # If the user enters an invalid expression
            return (1, "Invalid command syntax")
        except (NameError, ValueError) as err:
            # If the user tries to use a name that isn't allowed
            # or an invalid value for a given math function
            return (1, err)

        vinfo = self.parent.getVinfo(vnam)
        if vinfo is not None:
            nelem = self.parent.blklst[vinfo.blkpos][vinfo.vidx].size
            try:
                rsiz = R.size
            except AttributeError:
                # R is not a vector
                if isNumber(R):
                    self.parent.blklst[vinfo.blkpos][vinfo.vidx] = np.full(nelem, R)
                else:
                    return (1, "Invalid command syntax")
            if rsiz == nelem:
                self.parent.blklst[vinfo.blkpos][vinfo.vidx] = R
            else:
                return (1, "Vector size do not match")
        else:
            # If the vector does not exist, check if it is in expression
            if expression.find(vnam) != -1 or expression.find("("+vnam+")") != -1:
                return (1, "Unknown vector name")
            # Add a new vector at the end of a data block
            if blkno == -1:
                if not hasattr(R, "__len__"):
                    # R is not a vector
                    return (1, "Cannot process this command")
                # Look for a data block size matching R.size
                for i in range(len(self.parent.blklst)):
                    if self.parent.blklst[i][0].size == R.size:
                        blkno = i
                        break
            if blkno == -1:
                # No block size match R.size
                return (1, "Vector size do not match")
            s = np.shape(self.parent.blklst[blkno])
            nlin = s[0]
            if len(s) > 1:
                nrow = s[1]
            else:
                nrow = nlin
                nlin = 1
            vinfo = vectInfo(blkno, nlin, vnam)
            # add the new vector
            self.parent.vectInfolst[blkno].append(vinfo)
            newvect = R
            newset = np.vstack((self.parent.blklst[blkno], newvect))  # vstack = data stacked in row
            self.parent.blklst[blkno] = newset
        self.parent.dirty = True
        return (0, "")


    def exeNoArg(self, cmd):
        """ Process the commands with no argument

        :param cmd: a string containing the command
        :return: a tuple (errno, errmsg).
        """
        # look for the command name
        errno = 0
        msg = ""
        try:
            cmdidx = self.typ0Lst.index(cmd)
        except ValueError:
            return (1, "Check the parameters")
        logfilnam = os.path.join(self.parent.progpath, "logfile.txt")
        if cmdidx == self.typ0Lst.index('BET'):
            errno, msg = BET(self.parent.blklst[0], 'N2', logfilnam)
        elif cmdidx == self.typ0Lst.index('BETKr'):
            errno, msg = BET(self.parent.blklst[0], 'Kr', logfilnam)
        elif cmdidx == self.typ0Lst.index('BETAr'):
            errno, msg = BET(self.parent.blklst, 'Ar', logfilnam)

        elif cmdidx == self.typ0Lst.index('IRabs'):
            errno, msg = convertToAbs(self.parent.blklst[0])
            if not errno:
                self.parent.laby1 = "Absorbance"
                self.parent.vectInfolst[0][1].name = 'Absorbance'
        elif cmdidx == self.typ0Lst.index('IRtrans'):
            errno, msg = convertToTrans(self.parent.blklst[0])
            if not errno:
                self.parent.laby1 = "Transmittance"
                self.parent.vectInfolst[0][1].name = 'Transmittance'
        elif cmdidx == self.typ0Lst.index('invertx'):
                self.parent.invertX = not(self.parent.invertX)
                return errno, msg
        elif cmdidx == self.typ0Lst.index('inverty'):
                self.parent.invertY = not(self.parent.invertY)
                return errno, msg
        else:
            return (1, "Internal error")
        if errno == 0:
            if cmdidx > 3:       # cmdidx <= 3 for BET and PSD
                self.parent.dirty = True
        return (errno, msg)


    def exeOneArg(self, cmdidx, vnam):
        """ Process the commands with one argument

        :param cmdidx: index of the command in self.typ1Lst
        :param vnam: name of the vector argument.
        :return: a tuple (errno, errmsg).
        """
        if not isNumber(vnam):
            vinfo = self.parent.getVinfo(vnam)
            blkno = vinfo.blkpos
            vpos = vinfo.vidx
        else:
            blkno = None
            vpos = None

        if cmdidx == self.typ1Lst.index('stats'):
            err, result = stats(self.parent, vnam)
            return (0, result)

        if cmdidx == self.typ1Lst.index('area'):
            # check that the range is defined
            if len(self.parent.markList) == 1:
                return (1, "Range must be defined (with markers)")
            else:
                curvinfo = self.parent.getCurvinfo(vnam)
                if curvinfo is None:
                    return (1, "Unknown curve")
                if len(self.parent.markList) == 0:
                    idx1 = None
                    idx2 = None
                else:
                    idx1 = self.parent.markList[0].getIndex()
                    idx2 = self.parent.markList[1].getIndex()
                result = calculArea(self.parent, curvinfo, idx1, idx2)
                return (0, result)

        if cmdidx == self.typ1Lst.index('lineq'):
            if len(self.parent.markList) != 2:
                return (1, "Range must be defined (with markers)")
            else:
                curvinfo = self.parent.getCurvinfo(vnam)
                if curvinfo is None:
                    return (1, "Wrong curve name")
                idx1 = self.parent.markList[0].getIndex()
                idx2 = self.parent.markList[1].getIndex()
                return self.linEq(self.parent, curvinfo, idx1, idx2)

        if cmdidx == self.typ1Lst.index('delb'):
            blkno = int(vnam) - 1
            if self.parent.delBlock(blkno):
                return (0, "")
            else:
                return (1, "Cannot delete this data block")

        if cmdidx == self.typ1Lst.index('delv'):
            if self.parent.delVector(vnam):
                return (0, "")
            else:
                return (1, "Cannot delete this vector")

        if cmdidx == self.typ1Lst.index('line'):
            # Join by a straight line the mark positions
            err, errmsg = line(self.parent, vnam)
            if not err:
                self.parent.dirty = True
                self.parent.clearMarks()
            return (err, errmsg)

        if cmdidx == self.typ1Lst.index('revertb'):
            # Revert data order in data block
            blkno = int(vnam) - 1
            err, errmsg = revertb(self.parent, blkno)
            if not err:
                self.parent.dirty = True
                self.parent.clearMarks()
                errmsg = "The data order in all the vectors in block {0}" \
                         " has been reverted".format(blkno+1)
                err = -1
            return (err, errmsg)

        if cmdidx == self.typ1Lst.index('revertv'):
            # Revert data order in vnam
            err, errmsg = revertv(self.parent, vnam)
            if not err:
                self.parent.dirty = True
                self.parent.clearMarks()
                errmsg = "The data order in the vector {0}" \
                         " has been reverted".format(vnam)
                err = -1
            return (err, errmsg)

        if cmdidx == self.typ1Lst.index('sort'):
            # Sort vnam in ascending order.
            err, errmsg = sort(self.parent, vnam)
            if not err:
                self.parent.dirty = True
                self.parent.clearMarks()
                errmsg = "The vector {0} has been sorted" \
                         " in ascending order".format(vnam)
                err = -1
            return (err, errmsg)

        if cmdidx == self.typ1Lst.index('deldupx'):
            # Sort X in ascending order and remove duplicate values.
            err, errmsg = delDupX(self.parent, vnam)
            if not err:
                err = -1
                self.parent.dirty = True
                self.parent.clearMarks()
            return (err, errmsg)

        if cmdidx == self.typ1Lst.index('onset'):
            if len(self.parent.markList) != 2:
                return (1, "Positions must be defined (with markers)")
            else:
                curvinfo = self.parent.getCurvinfo(vnam)
                idx1 = self.parent.markList[0].getIndex()
                idx2 = self.parent.markList[1].getIndex()
                result = onset(self.parent, curvinfo, idx1, idx2)
                return (0, result)

        if cmdidx == self.typ1Lst.index('fft'):
            err, errmsg = fft(self.parent, vnam)
            return (err, errmsg)

        else:          # unknown command
            return (1, "Check the parameters")



    def exeTwoArg(self, cmd):
        """  Process the commands with two arguments.

             The commands clipup and clipdn are processed here
             although they have 3 arguments.

        :param cmd: a string containing the command name.
        :return: a tuple (errno, errmsg).
        """
        items = cmd.split(' ')
        cmdname = items[0].strip()
        # look for the command name
        try:
            cmdidx = self.typ2Lst.index(cmdname)
        except ValueError:
            return (1, "Check the parameters")

        if cmdname == 'mergeb':
            # Syntax is mergeb blk1 blk2
            err, msg = mergeb(self.parent, items[1:])
            if not err:
                self.parent.dirty = True
            return err, msg

        if cmdname == 'PSD':
            # Syntax is: PSD D halsey
            # use adsorption isotherm
            pos = 0
            if items[1].strip() == 'D':
                if len(self.parent.blklst) > 1:
                    # use desorption isotherm stored in the second data block.
                    pos = 1
                else:
                    # no adsorption isotherm only
                    # one desorption isotherm
                    pos = 0
            dirnam = os.path.dirname(self.parent.filename)
            basenam = os.path.basename(self.parent.filename)
            psdnam = basenam.replace("iso", "PSD")
            if psdnam == basenam:
                psdnam = "PSD.txt"
            psdnam = os.path.join(dirnam, psdnam)
            tcurv = items[2].strip()
            if tcurv in ['halsey', 'harkins', 'tfit']:
                errno, msg = PSDcalc(self.parent.blklst[pos], psdnam, tcurv)
            else:
                errno, msg = PSDcalc(self.parent.blklst[pos], psdnam)
            if not errno:
                msg = self.parent.parent.loadFile(psdnam)
            return 0, msg

        # The vector name is expected to be the 2nd argument (items[1))
        # except for the clipx command
        vnam = items[1].strip()
        if vnam != '<' and vnam != '>':
            # it not a clipx command
            vinfo = self.parent.getVinfo(vnam)
            if vinfo is None:
                return (1, "Unknown vector name")
            blkno = vinfo.blkpos
            vidx = vinfo.vidx

        # check value
        if isNumber(items[2]):
            val = float(items[2])
        else:
            v2nam = items[2].strip()

        if cmdname == 'swapv':
            # the syntax is: swapv V1 V2
            err, msg = swapv(self.parent, vnam, v2nam)
            if not err:
                self.parent.dirty = True
            return err, msg

        elif cmdname =='clipup':
            # the syntax is: clipup V > n
            err, msg = clipup(self.parent, vnam, val)
            if err and msg == "":
                err = 0
            else:
                self.parent.dirty = True
            return err, msg

        elif cmdname =='clipdn':
            # The syntax is: clipdn V > n
            err, msg = clipdn(self.parent, vnam, val)
            if err and msg == "":
                err = 0
            else:
                self.parent.dirty = True
            return err, msg

        elif cmdname =='clipx':
            # The syntax is: clipx < n
            err, msg = clipx(self.parent, val, vnam)
            if not err:
                self.parent.dirty = True
            return err, msg

        elif cmdname == 'linefit':
            # The syntax is: linefit V -1
            # check that the cursor position is defined
            if self.parent.dcursor is None:
                return (1, "Position must be defined (with cursor)")
            else:
                curvinfo = self.parent.getCurvinfo(vnam)
                idx = self.parent.dcursor.getIndex()
                err, result = lineFit(self.parent, curvinfo, idx, val)
                return err, result

        elif cmdname == 'shift':
            # The syntax is: shift V n
            if len(self.parent.markList) != 2:
                return (1, "Range must be defined (with markers)")
            else:
                curvinfo = self.parent.getCurvinfo(vnam)
                if curvinfo is None:
                    return (1, "Wrong curve name")
                idx1 = self.parent.markList[0].getIndex()
                idx2 = self.parent.markList[1].getIndex()
                err, msg = shift(self.parent, curvinfo, val, idx1, idx2)
                if not err:
                    self.parent.dirty = True
                return err, msg

        elif cmdname == 'shrink':
            # The syntax is: shrink V n
            factor = int(val)
            err, msg = shrink(self.parent, vnam, factor)
            if not err:
                self.parent.dirty = True
            return err, msg

        elif cmdname == 'despike':
            # The syntax is despike V winsiz
            winsiz = int(val)
            err, msg = despike(self.parent, vnam, winsiz)
            if not err:
                self.parent.dirty = True
            return err, msg

        elif cmdname == 'ndec':
            # Syntax is: ndec V n
            nd = int(items[2])
            err, msg = ndec(self.parent, vnam, nd)
            if not err:
                msg = "The number of decimal place of vector {0} was set to {1}".format(vnam, ndec)
                self.parent.dirty = True
            return err, msg

        else:
            return(1, "Unable to process command")



def name(pltw, cmd):
    """ Process a rename command

        :param pltw: plotWin containing data
        :param cmd: a string containing the command
        :return: a tuple (errno, errmsg).
    """
    # The syntax is: 'name oldnam as newnam'
    if cmd.find('as') == -1:
        return 1, "Check the parameters"
    cmd = cmd[5:]
    items = cmd.split('as')
    oldnam = items[0].strip()
    newnam = items[1].strip()
    vinfo = pltw.getVinfo(newnam)
    if vinfo is not None:
        return 1, "This name already exists"
    vinfo = pltw.getVinfo(oldnam)
    if vinfo is None:
        return 1, "Unknown vector name"
    vinfo.name = newnam
    # update pltw.curvelist
    for curvinfo in pltw.curvelist:
        if curvinfo.name == oldnam:
            curvinfo.name = newnam
    pltw.dirty = True
    return 0, ""


def newV(pltw, cmd):
    """ Process a new vector command

        :param pltw: plotWin containing data
        :param cmd: a string containing the command
        :return: a tuple (errno, errmsg).
    """
    errmsg = ""
    # The syntax is : newv V start,stop,step
    # Remove the cmommand name
    cmd = cmd[5:]

    items = cmd.split(' ')
    vname = items[0]
    n = len(vname)
    cmd = cmd[n:]
    items = cmd.split(',')
    # Check the user input
    if len(items) != 3:
        errmsg = "The syntax of this command is :\n" \
                 " Name Vini, Vend, step"
        return 1, errmsg

    # Check if a vector with the same name already exists
    if pltw.vectInfolst:
        vinfo = pltw.getVinfo(vname)
        if vinfo is not None:
            errmsg = "A vector {0} already exists".format(vname)
            return 1, errmsg

    # Check if Vini, Vend, step are numbers
    if isNumber(items[0]):
        start = float(items[0])
    else:
        errmsg = "Starting value must be a number"
        return 1, errmsg
    if isNumber(items[1]):
        stop = float(items[1])
    else:
        errmsg = "Ending value must be a number"
        return 1, errmsg
    if isNumber(items[2]):
        step = float(items[2])
    else:
        errmsg = "Step value must be a number"
        return 1, errmsg

    # Create the data list
    stop = stop + step
    newv = np.arange(start, stop, step)
    # Create a new data block for the new vector
    if pltw.blklst is None:
        pltw.blklst = []
    # add the new block at the end of block list
    pltw.blklst.append([newv])
    blkpos = len(pltw.blklst) - 1
    # Create the new vectInfo object
    vinfo = vectInfo(blkpos, 0, vname)
    # Add the new vectInfo to the list of vectInfo list
    vectInfolst = []
    vectInfolst.append(vinfo)
    pltw.vectInfolst.append(vectInfolst)
    pltw.dirty = True
    return 0, errmsg


def stats(pltw, vnam):
    """ Return statistics on vnam'.

        :param pltw: plotWin containing data
        :param vnam: the name of the relevant vector.
        :return: a tuple (err, msg) where err = 1 if an error occurred,
                 and msg contains the error message.
    """
    vinfo = pltw.getVinfo(vnam)
    if vinfo is None:
        return 1, "Unknown vector name"
    blkno = vinfo.blkpos
    vpos = vinfo.vidx
    result = "Statistics on {0}:\n".format(vnam)
    val = pltw.blklst[blkno][vpos].min()
    result += "   Mini = {0:g}\n".format(val)
    val = pltw.blklst[blkno][vpos].max()
    result += "   Maxi = {0:g}\n".format(val)
    val = val - pltw.blklst[blkno][vpos].min()
    val = val / (pltw.blklst[blkno][vpos].size - 1)
    dv = abs(pltw.blklst[blkno][vpos][1] - pltw.blklst[blkno][vpos][0])
    if dv == val:
        s = "constant"
    else:
        s = "average"
    result += "   Delta = {0:g},  {1}\n".format(val, s)
    val = pltw.blklst[blkno][vpos].sum()
    result += "   Sum = {0:g}\n".format(val)
    val = np.median(pltw.blklst[blkno][vpos])
    result += "   Median = {0:g}\n".format(val)
    val = pltw.blklst[blkno][vpos].mean()
    result += "   Mean = {0:g}\n".format(val)
    val = pltw.blklst[blkno][vpos].std()
    result += "   Variance = {0:g}\n".format(val)
    val = rms(pltw.blklst[blkno][vpos])
    result += "   Standard deviation = {0:g}\n\n".format(val)
    val = pltw.blklst[blkno][vpos].var()
    result += "Root mean square (RMS) = {0:g}\n".format(val)
    val = pp(pltw.blklst[blkno][vpos])
    result += "Peak-to-peak (pp) = {0:g}\n".format(val)
    return 0, result


def swapv(pltw, vnam1, vnam2):
    """ Exchange all elements between vectors named 'vnam1' and 'vnam2'.

        Of course these vectors must have the same size.
        Does not change the names.
        :param pltw: plotWin containing data
        :param vnam: the name of the relevant vector.
        :return: a tuple (err, msg) where err = 1 if an error occurred,
                 and msg contains the error message.
    """
    v1info = pltw.getVinfo(vnam1)
    if v1info is not None:
        v2info = pltw.getVinfo(vnam2)
    if v1info is None or v2info is None:
        return 1, "Unknown vector name"
    blkno1 = v1info.blkpos
    blkno2 = v2info.blkpos
    if blkno1 != blkno2:
        if pltw.blklst[blkno1][v1info.vidx].size != pltw.blklst[blkno2][v2info.vidx].size:
            return 1, "Both vectors must have the same size"
    tmp = pltw.blklst[blkno2][v2info.vidx].copy()
    pltw.blklst[blkno2][v2info.vidx] = pltw.blklst[blkno1][v1info.vidx]
    pltw.blklst[blkno1][v1info.vidx] = tmp
    return 0, ""


def shrink(pltw, vnam, n):
    """ Reduce in the number of elements in 'vnam' by a factor 'n'.

       :param pltw: plotWin containing data
       :param vnam: the name of the relevant vector.
       :return: a tuple (err, msg) where err = 1 if an error occurred,
                and msg contains the error message.
    """
    vinfo = pltw.getVinfo(vnam)
    if vinfo is None:
        return 1, "Unknown vector name"
    blkno = vinfo.blkpos
    if 1 < n < 10:
        ni = pltw.blklst[blkno][0].size
        pltw.blklst[blkno] = shrinkRows(pltw.blklst[blkno], n)
        nf = pltw.blklst[blkno][0].size
        msg = "The number of elements in {0} has been reduced " \
              "from {1} to {2}".format(vnam, ni, nf)
        return 0, msg
    else:
        return 1, "Shrink factor must be in range 2-10"


def despike(pltw, vnam, winsiz):
    """ Removes spikes, which point width are lower than winsiz, in 'vnam'.

       :param pltw: plotWin containing data
       :param vnam: the name of the relevant vector.
       :param winsiz: size, in number of points, of the largest spike to remove.
       :return: a tuple (err, msg) where err = 1 if an error occurred,
                and msg contains the error message.

    """
    vinfo = pltw.getVinfo(vnam)
    if vinfo is None:
        return 1, "Unknown vector name"
    blkno = vinfo.blkpos
    vidx = vinfo.vidx
    npt = pltw.blklst[0][vidx].size
    if isNumber(winsiz) and winsiz < 3 and (winsiz > int(npt / 10)):
        return 1, "Bad window size"
    if not winsiz % 2:
        return 1, "winsiz must be an odd integer"
    else:
        pltw.blklst[blkno][vidx] = signal.medfilt(pltw.blklst[blkno][vidx], winsiz)
        return 0, ""


def ndec(pltw, vnam, ndec):
    """ Set the number of decimal places to n for all 'vnam' elements.

        :param pltw: plotWin containing data
        :param vnam: the name of the relevant vector.
        :param ndec: the number of decimal places
        :return: a tuple (err, msg) where err = 1 if an error occurred,
                 and msg contains the error message.
    """
    vinfo = pltw.getVinfo(vnam)
    if vinfo is None:
        return 1, "Unknown vector name"
    blkno = vinfo.blkpos
    vidx = vinfo.vidx
    if isNumber(ndec) and ndec > 0 or ndec > 15:
        v = 10 ** ndec
        pltw.blklst[blkno][vidx] *= v
        pltw.blklst[blkno][vidx] = np.ceil(pltw.blklst[0][vidx])
        pltw.blklst[blkno][vidx] /= v
        return 0, ""
    else:
        return 1, "Invalid number of decimal"


def line(pltw, cnam):
    """ Join by a straight line the marker positions on the curve 'cnam'.

        :param pltw: plotWin containing data
        :param vnam: the name of the relevant curve.
        :return: a tuple (err, msg) where err = True if an error occurred,
                 and msg contains the error message.
    """
    if len(pltw.markList) != 2:
        return 1, "Need two marks to process"
    # check that vnam is a curve
    cinfo = pltw.getCurvinfo(cnam)
    if cinfo is None:
        return 1, "No curve matches with {0}".format(cnam)
    i1 = pltw.markList[0].getIndex()
    i2 = pltw.markList[1].getIndex()
    blkno = cinfo.yvinfo.blkpos
    X = pltw.blklst[blkno][cinfo.xvinfo.vidx]
    Y = pltw.blklst[blkno][cinfo.yvinfo.vidx]
    if X[i2] < X[i1]:
        i = i1
        i1 = i2
        i2 = i
    intercept = Y[i1]
    slope = (Y[i2] - Y[i1]) / (X[i2] - X[i1])
    for i in range(i1, i2):
        Y[i] = slope * (X[i] - X[i1]) + intercept
    return 0, ""


def revertb(pltw, blkno):
    """ Revert data order in the data block 'blkno'

        :param pltw: plotWin containing data
        :param blkno: the position of the data block which will be processed.
        :return: a tuple (err, msg) where err = True if an error occurred,
                 and msg contains the error message.
    """
    if blkno < 0 or blkno >= len(pltw.blklst):
        return 1, "Wrong block data number"
    pltw.blklst[blkno] = np.array([x[::-1] for x in pltw.blklst[blkno]])
    return 0, ""


def revertv(pltw, vnam):
    """ Revert data order in the vector 'vnam'

        :param pltw: plotWin containing data
        :param vnam: the name of the vector which will be processed.
        :return: a tuple (err, msg) where err = True if an error occurred,
                 and msg contains the error message.
    """
    errno = 0
    errmsg = ""
    vinfo = pltw.getVinfo(vnam)
    if vinfo is None:
        return 1, "Cannot find vector {0}".format(vnam)
    blkno = vinfo.blkpos
    vidx = vinfo.vidx
    pltw.blklst[blkno][vidx] = (pltw.blklst[blkno][vidx])[::-1]
    return errno, errmsg



def sort(pltw, vnam):
    """ Sort 'vnam' in ascending order.

        Process only the data block containing the vector 'vnam'.
        :param pltw: plotWin containing data
        :param vnam: name of the vector which will be sorted
        :return: a tuple (err, msg) where err = 1 if an error occurred,
                 and msg contains the error message.
    """
    vinfo = pltw.getVinfo(vnam)
    if vinfo is None:
        return 1, "Cannot find vector {0}".format(vnam)
    blkno = vinfo.blkpos
    vidx = vinfo.vidx
    pltw.blklst[blkno] = sortArr(pltw.blklst[blkno], vidx)
    return 0, ""


def delDupX(pltw, Xnam):
    """ First sort 'Xnam' in ascending order and then remove duplicates.

        Process only the data block containing the vector 'Xnam'.
        :param pltw: plotWin containing data
        'Xnam' must be the first vector of the data block.
        :param Xnam: the name of the X vector.
        :return: a tuple (err, msg) where err = 1 if an error occurred,
                 and msg contains the error message.
    """
    vinfo = pltw.getVinfo(Xnam)
    if vinfo is None:
        return 1, "Cannot find vector {0}".format(Xnam)
    blkno = vinfo.blkpos
    vidx = vinfo.vidx
    if vidx != 0:
        return 1, "{0} must be the first vector of the data block".format(Xnam)
    ni = pltw.blklst[blkno][0].size
    pltw.blklst[blkno] = delDupliX(pltw.blklst[blkno], vidx)
    nf = pltw.blklst[blkno][0].size
    dn = ni - nf
    if dn == 0:
        msg = "No duplicate elements has been found in {0}".format(Xnam)
    else:
        msg = "{0} duplicate elements has been deleted in {1}".format(dn, Xnam)
    return 0, msg


def clipup(pltw, vnam, val):
    """ Replace in the vector named 'vnam' all elements larger than val by val.

        :param pltw: plotWin containing data
        :param vnam: name of the relevant vector
        :param val: limit value
        :return: a tuple error value (=0 if no error) and error message
                 (="" if no error). If 'val' is higher than the maximum,
                 return err=1 and msg="" to indicate that nothing was done.
    """
    vinfo = pltw.getVinfo(vnam)
    if vinfo is None:
        return 1, "Cannot find vector {0}".format(vnam)
    blkno = vinfo.blkpos
    vidx = vinfo.vidx
    vm = np.amax(pltw.blklst[blkno][vidx])
    if vm <= val:
        return 1, ""
    vm = np.amin(pltw.blklst[blkno][vidx])
    pltw.blklst[blkno][vidx] = np.clip(pltw.blklst[blkno][vidx], vm, val)
    return 0, ""


def clipdn(pltw, vnam, val):
    """ Replace in the vector named 'vnam' all elements lower than val by val.

        :param pltw: plotWin containing data
        :param vnam: name of the relevant vector
        :param val: limit value
        :return: a tuple error value (=0 if no error) and error message
                 (="" if no error). If 'val' is lower than the minimum,
                 return err=1 and msg="" to indicate that nothing was done.
    """
    vinfo = pltw.getVinfo(vnam)
    if vinfo is None:
        return 1, "Cannot find vector {0}".format(vnam)
    blkno = vinfo.blkpos
    vidx = vinfo.vidx
    vm = np.amin(pltw.blklst[blkno][vidx])
    if vm >= val:
        return 1, ""
    vm = np.amax(pltw.blklst[blkno][vidx])
    pltw.blklst[blkno][vidx] = np.clip(pltw.blklst[blkno][vidx], val, vm)
    return 0, ""


def clipx(pltw, val, way):
    """ Remove all data points for which x is larger or lower than val.

        Data points for which x=val are kept.
        :param pltw: plotWin containing data
        :param val: limit value
        :param way: either '>' or '<'
        :return: an error value (=0 if no error) and an error message
                 (="" if no error). Return err=1 and msg="" to indicate
                 that nothing was done.
    """
    if not (way == '>' or way == '<'):
        return 1, "Bad parameter value"
    blkno = pltw.curvelist[0].yvinfo.blkpos
    xpos = pltw.curvelist[0].xvinfo.vidx
    xmin = pltw.blklst[blkno][xpos].min()
    xmax = pltw.blklst[blkno][xpos].max()
    if (way == '>' and val >= xmax) or (way == '<' and val <= xmin):
        # nothing to do, return
        return 0, ""
    # build a list of tuples containing unique (blkno, xpos) pair
    setlist = [(blkno, xpos)]
    ncurv = len(pltw.curvelist)
    for i in range(ncurv - 1):
        blkno = pltw.curvelist[i + 1].yvinfo.blkpos
        xpos = pltw.curvelist[i + 1].xvinfo.vidx
        if not (blkno, xpos) in setlist:
            setlist.append((blkno, xpos))
    nblk = len(setlist)
    # for each data block in setlist find either minidx or maxindx
    for i in range(nblk):
        blkno = setlist[i][0]
        xpos = setlist[i][1]
        (nvec, npt) = pltw.blklst[blkno].shape
        if way == '<':
            xmin = pltw.blklst[blkno][xpos].min()
            if val <= xmin:
                minidx = 0
            else:
                minidx = pltw.xToIdx(val, blkno, xpos)
                if minidx == None:
                    minidx = 0
            if pltw.xascending:
                newarr = np.delete(pltw.blklst[blkno], np.s_[:minidx], axis=1)
            else:
                newarr = np.delete(pltw.blklst[blkno], np.s_[minidx:], axis=1)
            pltw.blklst[blkno] = newarr
        else:  # way == '>'
            maxidx = pltw.xToIdx(val, blkno, xpos)
            if maxidx == None:
                maxidx = npt - 1
            else:
                if maxidx < npt - 2:
                    maxidx += 1
            if pltw.xascending:
                newarr = np.delete(pltw.blklst[blkno], np.s_[maxidx:], axis=1)
            else:
                newarr = np.delete(pltw.blklst[blkno], np.s_[:maxidx], axis=1)
            pltw.blklst[blkno] = newarr
    pltw.clearMarks()
    return 0, ""


def onset(pltw, curvinfo, idx1, idx2):
    """ Compute the onset of a peak.

        :param pltw: plotWin containing data
        :param curvinfo: vinfo object of the relevant curve.
        :param idx1: index of the point located on the baseline
        :param idx2: index of the point located on the peak
        :return: a tuple err, string where string contains the result
                 or an error message if err != 0.
    """
    # Algorithm
    # 1 - call linFit with direc=-1 to get the line equation at idx1
    # 2 - call linFit with direc=+1 to get the line equation at idx2
    # 3 - calculate the coordinates of the intersection of these lines
    # 4 - return the coordinates

    # If idx2 > idx1 swap them
    if idx2 < idx1:
        idx = idx1
        idx1 = idx2
        idx2 = idx
    (errmsg, pcoef1, n1, chi) = linFitProc(pltw, curvinfo, idx1, -1)
    if errmsg is None:
        (errmsg, pcoef2, n2, chi) = linFitProc(pltw, curvinfo, idx2, +1)
    if errmsg is not None:
        return 1, "Unable to compute the line equation"

    blkno = curvinfo.yvinfo.blkpos
    xpos = curvinfo.xvinfo.vidx
    ypos = curvinfo.yvinfo.vidx
    xi = (pcoef2[1] - pcoef1[1]) / (pcoef1[0] - pcoef2[0])
    yi = pcoef1[0] * xi + pcoef1[1]
    result = "The two lines meet at:\n"
    result += "xi = {0:g}\n".format(xi)
    result += "yi = {0:g}".format(yi)

    # Show the fitting lines
    siz = pltw.blklst[blkno][xpos].size
    ymin = pltw.blklst[blkno][ypos].min()
    ymax = pltw.blklst[blkno][ypos].max()
    # add the first line segment marker
    Yfit = np.polyval(pcoef1, pltw.blklst[blkno][xpos])
    x1 = pltw.blklst[blkno][xpos][idx1 - n1]
    y1 = Yfit[idx1 - n1]
    x2 = pltw.blklst[blkno][xpos][idx2]
    y2 = Yfit[idx2]
    pltw.lineList.append(lineSeg(x1, y1, x2, y2, 'red'))

    # add the second line segment marker
    Yfit = np.polyval(pcoef2, pltw.blklst[blkno][xpos])
    idx = idx1
    while Yfit[idx] < ymin and idx < idx2 - 1:
        idx += 1
    x1 = pltw.blklst[blkno][xpos][idx]
    y1 = Yfit[idx]
    x2 = pltw.blklst[blkno][xpos][idx2 + n2]
    y2 = Yfit[idx2 + n2]
    pltw.lineList.append(lineSeg(x1, y1, x2, y2, 'red'))

    pltw.displayInfo()
    pltw.plotCurves()
    return 0, result


def linEq(pltw, curvinfo, idx1, idx2):
    """ Compute the equation of the line linking two points on a curve.

        :param pltw: plotWin containing data
        :param curvinfo: vinfo object of the relevant curve.
        :param idx1: index of the first point
        :param idx2: index of the second point
        :return: a tuple (err, msg). If err=1, msg contains an error message
                 if err=0, msg contains the result of fitting.
    """
    blkno = curvinfo.yvinfo.blkpos
    xpos = curvinfo.xvinfo.vidx
    ypos = curvinfo.yvinfo.vidx
    X = pltw.blklst[blkno][xpos]
    Y = pltw.blklst[blkno][ypos]
    xm1 = pltw.markList[0].getxy()[0]
    idx1 = pltw.xToIdx(xm1, blkno, xpos)
    xm2 = pltw.markList[1].getxy()[0]
    idx2 = pltw.xToIdx(xm2, blkno, xpos)
    if idx1 < 0 or idx2 < 0 or idx1 >= X.size or idx2 >= X.size:
        return 1, "Index out of range"
    slope = (Y[idx2] - Y[idx1]) / (X[idx2] - X[idx1])
    intercept = Y[idx1] - slope * X[idx1]
    result = "Slope = {0:g}\nIntercept = {1:g}".format(slope, intercept)
    return 0, result


def shift(pltw, curvinfo, val, idx1=None, idx2=None):
    """ Add the scalar 'val' to a curve in the location defined by idx1, idx2.

        If idx1 and idx2 are None use the marker positions.
        :param pltw: plotWin containing data
        :param curvinfo: vinfo object of the relevant curve.
        :param val: the scalar value to add.
        :param idx1: the index of the first point
        :param idx2: the index of the last point
        :return: (0,"")
        """
    blkno = curvinfo.yvinfo.blkpos
    xpos = curvinfo.xvinfo.vidx
    ypos = curvinfo.yvinfo.vidx
    if idx1 is None:
        xm1 = pltw.markList[0].getxy()[0]
        idx1 = pltw.xToIdx(xm1, blkno, xpos)
    if idx2 is None:
        xm2 = pltw.markList[1].getxy()[0]
        idx2 = pltw.xToIdx(xm2, blkno, xpos)
    npt = pltw.blklst[blkno][xpos].size
    if idx1 < 0 or idx2 < 0 or idx1 >= npt or idx2 >= npt:
        return 1, "Index out of range"
    if idx1 != idx2:
        if idx1 > idx2:
            idx = idx2
            idx2 = idx1
            idx1 = idx
        for i in range(idx1, idx2 + 1):
            pltw.blklst[blkno][ypos][i] += val
    return 0, ""


def lineFit(pltw, curvinfo, idx, direc):
    """ Compute the equation of the line fitting a curve

        :param pltw: plotWin containing data
        :param curvinfo: vinfo object of the relevant curve.
        :param idx: index of the starting point.
        :param direc: must be either +1 or -1
        :return: a tuple (err, msg). If err=True, msg contains an error message
                 if err=False, msg contains the result of fitting.
    """
    blkno = curvinfo.yvinfo.blkpos
    xpos = curvinfo.xvinfo.vidx
    ypos = curvinfo.yvinfo.vidx
    ynam = curvinfo.name
    err = True
    (errmsg, pcoef, npt, chi) = linFitProc(pltw, curvinfo, idx, direc)
    if errmsg is None:
        err = False
        result = "Fitting curve {0} with a line\n".format(ynam)
        result += "Using {0:d} points\n".format(npt)
        result += "chi = {0:g}\n".format(chi)
        result += "intercept = {0:g}\nslope = {1:g}".format(pcoef[1], pcoef[0])
        # add a line segment marker to show the result
        siz = pltw.blklst[blkno][xpos].size
        ymin = pltw.blklst[blkno][ypos].min()
        ymax = pltw.blklst[blkno][ypos].max()
        Yfit = np.polyval(pcoef, pltw.blklst[blkno][xpos])
        if direc > 0:
            x1 = pltw.blklst[blkno][xpos][idx]
            y1 = Yfit[idx]
            end = idx + npt
            if end >= siz:
                end = siz - 1
            y2 = Yfit[end]
            while y2 < ymin and end > idx + 1:
                end -= 1
                y2 = Yfit[end]
            x2 = pltw.blklst[blkno][xpos][end]
        else:
            x2 = pltw.blklst[blkno][xpos][idx]
            y2 = Yfit[idx]
            ini = idx - npt
            if ini <= 0:
                ini = 0
            y1 = Yfit[ini]
            while y1 > ymax and ini < idx - 1:
                ini += 1
                y1 = Yfit[ini]
            x1 = pltw.blklst[blkno][xpos][ini]
        pltw.lineList.append(lineSeg(x1, y1, x2, y2, 'red'))
        pltw.displayInfo()
        pltw.plotCurves()
    else:
        result = errmsg
    return err, result


def linFitProc(pltw, curvinfo, idx, direc):
    """ Compute the equation of the line fitting a curve.

        :param pltw: plotWin containing data
        :param curvinfo: a curveInfo object containing the curve information.
        :param idx: the index of the point from which the fitting will start.
        :param direc: = 1 if the fitting proceed toward increasing X or equal
                      to -1 otherwise.
        :return: a tuple (errmsg, pcoef, npt, chi) where:
             - 'pcoef' contains the line parameters (intercept and slope).
             - 'npt' is the number of point used for fitting
             If no error 'errmsg' is None otherwise it contains an error message.
    """
    # Algorithm
    # 1 - Starts, in the direction given by 'direc', from 'idx' position
    #     with at least 5 points
    # 2 - Fit with a line
    # 3 - Add another point in the direction given by 'direc'
    # 4 - Return to step 2
    # Stop when enough points are used or if the fitting error exceeds
    # a given limit
    #
    errmsg = None
    if direc > 0:
        direc = 1
    else:
        direc = -1
    blkno = curvinfo.yvinfo.blkpos
    xpos = curvinfo.xvinfo.vidx
    ypos = curvinfo.yvinfo.vidx
    siz = pltw.blklst[blkno][xpos].size
    initrange = 5
    # Evaluate the maximum amplitude
    ymin = pltw.blklst[blkno][ypos].min()
    ymax = pltw.blklst[blkno][ypos].max()
    std = abs(ymax - ymin)
    if direc > 0:
        nstep = siz - idx - initrange - 1
    else:
        nstep = idx - initrange - 1
        if nstep < 0:
            nstep = 0
    chi = None
    pcoef = None
    tol = 0.002
    n = initrange
    for i in range(nstep):
        if direc > 0:
            l1 = idx
            l2 = idx + n
        else:
            l1 = idx - n
            l2 = idx
        X = pltw.blklst[blkno][xpos][l1:l2]
        Y = pltw.blklst[blkno][ypos][l1:l2]
        pcoef = np.polyfit(X, Y, 1)
        Yfit = np.polyval(pcoef, X)
        # compute the mean square error
        chi = np.sqrt(sum((Yfit - Y) ** 2) / len(Y))
        if chi / std > tol:
            break
        n += 1
    if chi is None:
        errmsg = "Unable to compute the line equation"
    return errmsg, pcoef, n, chi


def fft(pltw, vnam):
    """ Compute frequency spectrum using FFT.

        :param pltw: plotWin containing data
        :param vnam: name of the vector which will be used.
        :return: an error value (=0 is no error) and a message.
    """
    # http://glowingpython.blogspot.fr/2011/08/how-to-plot-frequency-spectrum-with.html
    # Recording of a time serie of N signals for a total time T,
    # dt = T/N
    # T and dt are the main properties of Fourier analysis.
    # A signal observed for a total time T allows a frequency resolution equal to 1/T
    # A signal observed with a sampling dt allows to get the frequencies up to 1(2.dt).
    #
    err = 1
    if pltw.parent.progpath is None:
        return "Cannot create FFT file"
    curvinfo = pltw.getCurvinfo(vnam)
    if curvinfo is None:
        return err, "No vector to process"
    Y = pltw.blklst[curvinfo.yvinfo.blkpos][curvinfo.yvinfo.vidx]
    # fft computing and normalization
    from numpy.fft import fft
    n = Y.size
    FFT = fft(Y) / n
    m = int(n / 2)
    FFT = FFT[range(m)]
    FFT = abs(FFT)
    k = np.arange(n)
    Fs = n
    T = n / Fs
    frq = k / T  # two sides frequency range
    frq = frq[range(m)]  # one side frequency range
    # Save FFT in a file
    filnam = "{0}-fft.txt".format(vnam)
    filnam = os.path.join(pltw.parent.progpath, filnam)
    try:
        np.savetxt(filnam, np.transpose([frq, FFT]), delimiter='\t')
    except PermissionError as err:
        return err, "Fail to create FFT file: {0}".format(err)
    msg = pltw.parent.loadFile(filnam)
    if msg:
        return err, "Cannot load {0}\n{1}".format(filnam, msg)
    else:
        return 0, msg



def mergeb(pltw, items):
    """ Merge two data blocks

        The data block vectors must have the same size.

    :param pltw: plotWin containing data
    :param items: list containing block numbers
    :return: a tuple (err, errmsg)
    """
    # check the parameters
    if len(items) > 2:
        return 1, "Only two blocks can be merged"
    if len(items) < 2:
        return 1, "mergeb command needs two block number"
    nb = len(pltw.blklst)
    if isNumber(items[0]):
        nb1 = int(items[0])
    if isNumber(items[1]):
        nb2 = int(items[1])
    if nb1 < 1 or nb1 > nb or nb2 < 1 or nb2 > nb:
        return 1, "Incorrect parameters"
    nb1 -= 1
    nb2 -= 1
    # check vector size
    if pltw.blklst[nb1][0].size != pltw.blklst[nb2][0].size:
        return 1, "The vectors must have the same size in both data blocks"
    nvect, nelem = pltw.blklst[nb2].shape
    for i in range(nvect):
        pltw.blklst[nb1] = np.vstack((pltw.blklst[nb1], pltw.blklst[nb2][i]))
    del pltw.blklst[nb2]
    idxshift = len(pltw.vectInfolst[nb1])
    # update pltw.vectInfolst
    for vinfo in pltw.vectInfolst[nb2]:
        vinfo.blkpos = nb1
        vinfo.vidx += idxshift
        pltw.vectInfolst[nb1].append(vinfo)
    del pltw.vectInfolst[nb2]
    # update pltw.curveInfolst
    for cinfo in pltw.curvelist:
        for vinfo in pltw.vectInfolst[nb1]:
            if cinfo.xvinfo.name == vinfo.name:
                cinfo.xvinfo = vinfo
            if cinfo.yvinfo.name == vinfo.name:
                cinfo.yvinfo = vinfo
    return 0, ""


def calculArea(plotw, curvinfo, idx1=None, idx2=None):
    """ Compute the area between a curve and x-axis.

        :param plotw: plotWin containing data
        :param curvinfo: vinfo object of the relevant curve.
        :param idx1: index of the first point
        :param idx2: index of the last point
        :return: a string containing the result.
    """
    if curvinfo is None:
        return ""
    link = True
    blkno = curvinfo.yvinfo.blkpos
    xpos = curvinfo.xvinfo.vidx
    ypos = curvinfo.yvinfo.vidx
    xnam = curvinfo.xvinfo.name
    ynam = curvinfo.yvinfo.name
    maxidx = len(plotw.blklst[blkno][xpos]) - 1
    if idx1 is None:
        idx1 = 0
    if idx2 is None:
        idx2 = maxidx
    if plotw.blklst[blkno][xpos][idx1] < plotw.blklst[blkno][xpos][idx2]:
        lowidx = idx1
        upidx = idx2
    else:
        lowidx = idx2
        upidx = idx1
    # remove data outside the limits
    dellst = list(range(0, lowidx)) + list(range(upidx + 1, maxidx + 1))
    x = np.delete(plotw.blklst[blkno][xpos], dellst)
    y = np.delete(plotw.blklst[blkno][ypos], dellst)
    a = calcArea(x, y)
    lowlim = plotw.blklst[blkno][xpos][lowidx]
    uplim = plotw.blklst[blkno][xpos][upidx]
    result = "Area between {0:g} and {1:g} = {2:g}\n\n".format(lowlim, uplim, a)
    result += "This area is defined by the curve {0}=f({1})".format(ynam, xnam)
    result += " and the x-axis"
    return result







import os
from time import sleep
import numpy as np
from PyQt5 import QtGui, QtWidgets

from pyDataVis.plotWindow import dCursor, lineSeg
from pyDataVis.script import calculArea, clipdn, clipup, clipx, delDupX, despike, fft
from pyDataVis.script import line, linEq, lineFit, mergeb, name, ndec, newV, onset
from pyDataVis.script import revertb, revertv, shift, shrink, sort, stats, swapv
from pyDataVis.utils import textToData
from pyDataVis.convIso import processBET, PSDcalc
from pyDataVis.convCIF import CIF
from pyDataVis.convert import convDAT, convPRF, convertShape, transpose, convert2D
from pyDataVis.convert import convertToAbs, convertToTrans

# - testDlg class ------------------------------------------------------------

class testDlg(QtWidgets.QDialog):
    def __init__(self, parent=None, group=None):
        """ Test dialog.

        :param parent: pyDataVis MainWindow.
        :param group: class of the tests ("script", "IO", "Convert")
        """
        super(testDlg, self).__init__(parent)
        info = "Test information is shown in the Text editor Window\n\n"
        info += "Hit the space bar to pause the testing process.\n"
        info += "Hit the Esc key to cancel the testing process.\n\n"
        info += "Select the test"
        infolab = QtWidgets.QLabel(info)
        self.testComboBox = QtWidgets.QComboBox(self)
        self.testlist = ["All tests"]
        if group == "script":
            self.setWindowTitle('Tests of script commands')
            self.testlist.extend(sctstnam)
        if group == "IO":
            self.setWindowTitle('Tests of Input/Output')
            self.testlist.extend(iotstnam)
        if group == "Convert":
            self.setWindowTitle('Tests of convert options')
            self.testlist.extend(convtstnam)
        self.testComboBox.addItems(self.testlist)
        # Buttons
        okBtn = QtWidgets.QPushButton("OK")
        cancelBtn = QtWidgets.QPushButton("Cancel")
        # set the layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(infolab)
        vbox.addWidget(self.testComboBox)
        hbox = QtWidgets.QHBoxLayout()
        hbox.addWidget(okBtn)
        hbox.addWidget(cancelBtn)
        vbox.addLayout(hbox)
        self.setLayout(vbox)

        # Connect buttons to callback functions
        okBtn.clicked.connect(self.accept)
        cancelBtn.clicked.connect(self.reject)


def wait(tim, mainw):
    """ Wait for 'tim' seconds

    :param mainw: the application MainWindow
    :param tim: the number of seconds to wait.
    :return: False if the user has pressed the Esc key.
    """
    for i in range(tim*10):
       if mainw.keyb == "escape":
           mainw.keyb = None
           return False
       if mainw.keyb == " ":
           mainw.keyb = None
           msgBox = QtWidgets.QMessageBox()
           msgBox.setIcon(QtWidgets.QMessageBox.Information)
           msgBox.setText("Ok to resume or Cancel to cancel")
           msgBox.setWindowTitle("Testing process paused")
           msgBox.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
           if msgBox.exec() == QtWidgets.QMessageBox.Cancel:
               return False
       sleep(0.1)
       mainw.app.processEvents()
    return True


def runTests(mainw, sel):
    """ Run the self tests

    :param mainw:
    :return: Nothing
    """
    if sel == "Run script tests":
        group = "script"
        tstfunc = sctstfunc
        tstnam = sctstnam
        title = "Script Tests results"
    elif sel == "Run I/O tests":
        group = "IO"
        tstfunc = iotstfunc
        tstnam = iotstnam
        title = "Input/Output Tests results"
    elif sel == "Run convert tests":
        group = "Convert"
        tstfunc = convtstfunc
        tstnam = convtstnam
        title = "Convert Tests results"
    else:
        return
    dlg = testDlg(mainw, group)
    dlg.setModal(True)
    ret = dlg.exec_()
    if not ret:
        return

    mainw.keyb = None
    pos = dlg.testComboBox.currentIndex()
    if pos == 0:
        # run all tests
        result = []
        for i, func in enumerate(tstfunc):
            err, msg = func(mainw)
            if err == -1:
                # user interruption
                break
            if err == 0:
                res = "passed"
            else:
                res = "failed\n{0}".format(msg)
            result.append("{0} test {1}".format(tstnam[i], res))
        result = '\n'.join(result)
    else:
        err, msg = tstfunc[pos - 1](mainw, extratim=3)
        if err == -1:
            res = "stopped by user"
        elif err == 0:
            res = "passed"
        else:
            res = "failed\n{0}".format(msg)
        result = "{0} test : {1}".format(tstnam[pos - 1], res)
    if result:
        QtWidgets.QMessageBox.information(mainw, title, result)


# ------------------------------------------------------------------------

def test_area(mainw, extratim=0):
    """ Test the 'area' script command.

        This also test Cursor and Marker
    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "area.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    err = 0
    pltw = subw.widget()
    pltw.dcursor = dCursor(pltw)
    if pltw.dcursor.move(indx=566):
        info = "#### Test of the 'area' script command ####\n"
        info += "# Test file: /selftests/area.plt\n"
        info += "# Set Cursor at X = 250"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    else:
        err = 1
        msg = "Data cursor error"
    if not err:
        mainw.addMark()
        info += "\n# Set a marker at cursor position"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    if not err:
        if pltw.dcursor.move(indx=900):
            info += "\n# Set Cursor at X = 362"
            pltw.datatext.setText(info)
            if not wait(2, mainw):
                err = -1
        else:
            err = 1
            msg = "Data cursor error"
    if not err:
        mainw.addMark()
        info += "\n# Set a marker at cursor position"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    if not err:
        # Area between 250.792 and 362.125 = 5562.28
        result = calculArea(pltw, pltw.curvelist[0], 566, 900)
        if result:
            info = "#\n# {0}\n\nEnd of the test".format(result)
            pltw.datatext.setText(info)
            if not wait(3 + extratim, mainw):
                err = -1
        else:
            err = 1
            msg = "Calcul area error"
    subw.close()
    return err, msg


def test_BET(mainw, extratim=0):
    """ Test the 'BET' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "BET.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    err = 0
    pltw = subw.widget()
    info = "#### Test of the 'BET' script command ####\n"
    info += "# Test file: /selftests/BET.plt\n"
    pltw.datatext.setText(info)
    if not wait(2, mainw):
        err = -1
    else:
        (msg, bestprm, Pri, Prf) = processBET(pltw.blklst[0])
        if msg:
            err = 1
    if not err:
        info += "# Computing BET surface area\n"
        info += '# BET Surface Area : {0:7.2f} m2/g\n'.format(bestprm.SBET)
        info += '# Correlation Coefficient :  {0:4.5f} \n\n'.format(bestprm.cc)
        info += "# End of the test"
        pltw.datatext.setText(info)
        if not wait(2 + extratim, mainw):
            err = -1
    subw.close()
    return err, msg


def test_clipupdn(mainw, extratim=0):
    """ Test the 'clipup' and 'clipdn' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "clipupdn.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    err = 0
    pltw = subw.widget()
    info = "#### Test of the 'clipup' and 'clipdn' script command ####\n"
    info += "# Test file: /selftests/clipupdn.plt\n\n"
    info += "# First, testing 'clipup' script command \n"
    pltw.datatext.setText(msg)
    if not wait(4, mainw):
        err = -1
    else:
        err, msg = clipup(pltw, 'Y', 1.0)
        if not err:
            pltw.updatePlot()
            info += "# Using 'clipup Y 1' to clip Y > 1.0\n"
            pltw.datatext.setText(info)
            if not wait(3, mainw):
                err = -1
    if not err:
        info = "# Now, testing 'clipdn' script command \n\n"
        info += "# Using 'clipdn Y -1' to clip Y < -1.0\n\n"
        info += "# End of the test"
        err, msg = clipdn(pltw, 'Y', -1.0)
        if not err:
            pltw.updatePlot()
            pltw.datatext.setText(info)
            if not wait(3 + extratim, mainw):
                err = -1
    subw.close()
    return err, msg


def test_clipx(mainw, extratim=0):
    """ Test the 'clipx' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "clipx.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    err = 0
    pltw = subw.widget()
    info = "#### Test of the 'clipx' script command ####\n"
    info += "# Test file: /selftests/clipx.plt\n\n"
    pltw.datatext.setText(info)
    if not wait(3, mainw):
        err = -1
    else:
        err, msg = clipx(pltw, 30, '>')
        if not err:
            pltw.updatePlot()
            info += "# Using 'clipx > 30' to remove X values > 30\n"
            info += "\n# End of the test"
            pltw.datatext.setText(info)
            if not wait(2 + extratim, mainw):
                err = -1
    subw.close()
    return err, msg


def test_delb(mainw, extratim=0):
    """ Test the 'delb' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "delb.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    err = 0
    pltw = subw.widget()
    info = "#### Test of the 'delb' script command ####\n"
    info += "# Test file: /selftests/delb.plt\n\n"
    info += "# This file contains 3 data blocks.\n"
    info += "# We will delete the block no 2."
    pltw.datatext.setText(info)
    if not wait(3, mainw):
        err = -1
    else:
        info += "\n# Executing 'delb 2' command\n"
        info += "# End of the test"
        pltw.datatext.setText(info)
        if pltw.delBlock(1):
            pltw.dirty = False
            if not wait(3 + extratim, mainw):
                err = -1
        else:
            err = 1
            msg = "Wrong block number"
    subw.close()
    return err, msg


def test_deldupx(mainw, extratim=0):
    """ Test the 'deldupx' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "deldupx.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Test of the 'deldupx' script command ####\n"
    info += "# Test file: /selftests/deldupx.plt\n\n"
    pltw.datatext.setText(info)
    if not wait(3, mainw):
        err = -1
    else:
        err, msg = delDupX(pltw, 'X')
    if not err:
        info += "# {0}\n\n# End of the test".format(msg)
        pltw.datatext.setText(info)
        if not wait(2 + extratim, mainw):
            err = -1
    subw.close()
    return err, msg


def test_delv(mainw, extratim=0):
    """ Test the 'delv' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "delv.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    err = 0
    pltw = subw.widget()
    info = "#### Test of the 'delv' script command ####\n"
    info += "# Test file: /selftests/delv.plt\n\n"
    pltw.datatext.setText(info)
    if not wait(2, mainw):
        err = -1
    else:
        info += "# Execute 'delv Pd' command\n\n"
        info += "# End of the test"
        pltw.datatext.setText(info)
        if pltw.delVector('Pd'):
            pltw.dirty = False
            if not wait(2 + extratim, mainw):
                err = -1
        else:
            err = 1
            msg = "Error in vector name"
    subw.close()
    return err, msg


def test_despike(mainw, extratim=0):
    """ Test the 'despike' script commands.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "despike.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Test of the 'despike' script command ####\n"
    info += "# Test file: /selftests/despike.plt\n\n"
    pltw.datatext.setText(info)
    if not wait(3, mainw):
        err = -1
    else:
        info += "# Execute 'despike I 3' command\n"
        info += "# End of the test"
        pltw.datatext.setText(info)
        err, msg = despike(pltw, 'I', 3)
    if not err:
        pltw.updatePlot()
        if not wait(3 + extratim, mainw):
            err = -1
    subw.close()
    return err, msg


def test_fft(mainw, extratim=0):
    """ Test the 'FFT' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "fft.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Test of the 'fft' script command ####\n"
    info += "# Test file: /selftests/fft.plt\n\n"
    pltw.datatext.setText(info)
    if not wait(1 + extratim, mainw):
        err = -1
    else:
        err, msg = fft(pltw, 'Y')
    if not err:
        info += "# Execute 'fft Y' command\n"
        info += "# This creates and loads the file Y-fft.txt\n\n"
        info += "# End of the test"
        pltw.datatext.setText(info)
        if not wait(3 + extratim, mainw):
            err = -1
        fftnam = "{0}/Y-fft.txt".format(mainw.progpath)
        subwfft = mainw.isAlreadyOpen(fftnam)
        if subwfft is not None:
            subwfft.close()
    subw.close()
    return err, msg


def test_IR(mainw, extratim=0):
    """ Test the 'IRtrans' and 'IRabs' script commands.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "IR.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Test of the 'IRabs' and 'IRtrans' script commands ####\n"
    info += "# Test file: /selftests/IR.plt\n\n"
    pltw.datatext.setText(info)
    if not wait(2, mainw):
        err = -1
    else:
        info = "##### First testing 'IRabs' script command #####\n\n"
        pltw.datatext.setText(info)
        err, msg = convertToAbs(pltw.blklst[0])
    if not err:
        pltw.updatePlot()
        if not wait(2 + extratim, mainw):
            err = -1
    if not err:
        info = "##### Now testing 'IRtrans' script command #####\n\n"
        info += "# End of the test"
        pltw.datatext.setText(info)
        err, msg = convertToTrans(pltw.blklst[0])
        if not err:
            pltw.updatePlot()
            if not wait(3 + extratim, mainw):
                err = -1
    subw.close()
    return err, msg


def test_line(mainw, extratim=0):
    """ Test the 'line' script command.

        This also test Cursor and Marker
    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "line.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    err = 0
    pltw = subw.widget()
    info = "#### Test of the 'line' script commands ####\n"
    info += "# Test file: /selftests/line.plt\n"
    pltw.datatext.setText(info)
    if not wait(2, mainw):
        err = -1
    pltw.dcursor = dCursor(pltw)
    if pltw.dcursor.move(indx=270):
        info += "# Set Cursor at X = 28.5\n"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    else:
        err = 1
        msg = "Data cursor error"
    if not err:
        mainw.addMark()
        info += "# Set a marker at cursor position\n"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    if not err:
        if pltw.dcursor.move(indx=414):
            info += "# Set Cursor at X = 33.0\n"
            pltw.datatext.setText(info)
            if not wait(2, mainw):
                err = -1
        else:
            err = 1
            msg = "Data cursor error"
    if not err:
        mainw.addMark()
        info += "# Set a marker at cursor position\n"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    if not err:
        err, ms = line(pltw, 'Y')
        if not err:
            pltw.updatePlot()
            info += "# Executing 'line Y' clears the second peak\n\n"
            info += "# End of the test"
            pltw.datatext.setText(info)
            if not wait(3 + extratim, mainw):
                err = -1
    subw.close()
    return err, msg


def test_linefit(mainw, extratim=0):
    """ Test the 'linefit' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "linefit.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    err = 0
    pltw = subw.widget()
    info = "#### Test of the 'linefit' script commands ####\n"
    info += "# Test file: /selftests/linefit.plt\n"
    pltw.datatext.setText(info)
    if not wait(3, mainw):
        err = -1
    else:
        pltw.dcursor = dCursor(pltw)
        if pltw.dcursor.move(indx=53):
            info += "# Set Cursor on the left side of the peak"
            pltw.datatext.setText(info)
            if not wait(3, mainw):
                err = -1
        else:
            err = 1
            msg = "Data cursor error"
    if not err:
        err, msg = lineFit(pltw, pltw.curvelist[0], 53, 1)
        if not err:
            info = "# Execute 'linefit Y' "
            info += "\n{}\n".format(msg)
            info += "\n# End of the test"
            pltw.datatext.setText(msg)
            if not wait(3 + extratim, mainw):
                err = -1
    subw.close()
    return err, msg


def test_lineq(mainw, extratim=0):
    """ Test the 'lineq' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "lineq.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    err = 0
    pltw = subw.widget()
    info = "#### Test of the 'lineq' script commands ####\n"
    info += "# Test file: /selftests/lineq.plt\n"
    pltw.datatext.setText(info)
    if not wait(2 + extratim, mainw):
        err = -1
    else:
        pltw.dcursor = dCursor(pltw)
        if pltw.dcursor.move(indx=20):
            info += "# Set Cursor at X = 1.0\n"
            pltw.datatext.setText(info)
            if not wait(2, mainw):
                err = -1
        else:
            err = 1
            msg = "Data cursor error"
    if not err:
        mainw.addMark()
        info += "# Set a marker at cursor position\n"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    if not err:
        if pltw.dcursor.move(indx=60):
            info += "# Set Cursor at X = 3.0\n"
            pltw.datatext.setText(info)
            if not wait(2, mainw):
                err = -1
        else:
            err = 1
            msg = "Data cursor error"
    if not err:
        mainw.addMark()
        info += "# Set a marker at cursor position\n"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    if not err:
        err, msg = linEq(pltw, pltw.curvelist[0], 20, 60)
        x1 = pltw.blklst[0][0][20]
        y1 = pltw.blklst[0][1][20]
        x2 = pltw.blklst[0][0][60]
        y2 = pltw.blklst[0][1][60]
        pltw.lineList.append(lineSeg(x1, y1, x2, y2, 'red'))
        pltw.updatePlot()
        if not err:
            info = "# Execute 'lineq Y' \n"
            info += "{0}\n\n# End of the test".format(msg)
            pltw.datatext.setText(info)
            if not wait(3 + extratim, mainw):
                err = -1
    subw.close()
    return err, msg


def test_mergeb(mainw, extratim=0):
    """ Test the 'mergeb' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "mergeb.txt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Test of the 'mergeb' script commands ####\n"
    info += "# Test file: /selftests/mergeb.txt\n\n"
    info += "# This file contains two data blocks with only one vector, \n"
    info += "# The 'mergeb 1 2' command will merge these blocks such as \n"
    info += "# there will be only one data block containing 2 vectors \n"
    pltw.datatext.setText(info)
    if not wait(4, mainw):
        err = -1
    else:
        err, msg = mergeb(pltw, [1, 2])
    if not err:
        tmpnam = "{0}/selftests/tmp.txt".format(mainw.testspath)
        done = pltw.save(tmpnam)
        if done:
            pltw.load(tmpnam)
            info += "# Execute 'mergeb 1 2' \n"
            info += "# End of the test"
            pltw.datatext.setText(info)
            if not wait(4 + extratim, mainw):
                err = -1
    subw.close()
    return err, msg


def test_ndec(mainw, extratim=0):
    """ Test the 'ndec' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "ndec.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    mainw.tableTool()
    info = "#### Test of the 'ndec' script commands ####\n"
    info += "# Test file: /selftests/ndec.plt\n"
    info += "# Display the data table"
    pltw.datatext.setText(info)
    if not wait(5, mainw):
        err = -1
    else:
        err, msg = ndec(pltw, 'Y', 2)
        pltw.tabledlg.table.initTable()
    if not err:
        info += "\n# Execute 'ndec Y 2' command"
        info += "\n# Now, the Y values are rounded to 2 decimal place"
        info += "\n\n# End of the test"
        pltw.datatext.setText(info)
        if not wait(5 + extratim, mainw):
            err = -1
        pltw.tabledlg.close()
    subw.close()
    return err, msg


def test_newv(mainw, extratim=0):
    """ Test the 'newv' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    mainw.fileNew()
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Test of the 'newv' script commands ####\n"
    pltw.datatext.setText(info)
    if not wait(1, mainw):
        err = -1
    else:
        err, msg = newV(pltw, "newv X 0,10,0.1")
    if not err:
        info += "# Executing 'newv X 0,10,0.1' command\n"
        info += "# This create a new vector X and\n"
        info += "# a new data block for hosting X\n"
        info += "\n# End of the test"
        pltw.datatext.setText(info)
        pltw.displayInfo()
    if not wait(4 + extratim, mainw):
        err = -1
    pltw.dirty = False
    subw.close()
    return err, msg


def test_onset(mainw, extratim=0):
    """ Test the 'linefit' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "onset.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Test of the 'onset' script commands ####\n"
    info += "# Test file: /selftests/onset.plt\n"
    pltw.datatext.setText(info)
    if not wait(2, mainw):
        err = -1
    else:
        err = 0
        pltw.dcursor = dCursor(pltw)
        if pltw.dcursor.move(indx=45):
            info += "# Set Cursor on the baseline\n"
            pltw.datatext.setText(info)
            if not wait(2, mainw):
                err = -1
        else:
            err = 1
            msg = "Data cursor error"
    if not err:
        mainw.addMark()
        info += "# Set a marker at cursor position\n"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    if not err:
        if pltw.dcursor.move(indx=52):
            info += "# Set Cursor on the left side of the peak\n"
            pltw.datatext.setText(info)
            if not wait(2, mainw):
                err = -1
        else:
            err = 1
            msg = "Data cursor error"
    if not err:
        mainw.addMark()
        info += "# Set a marker at cursor position\n"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    if not err:
        err, msg = onset(pltw, pltw.curvelist[0], 45, 52)
        if not err:
            info += "# Execute 'onset Y' command"
            info += "{0}\n# End of the test".format(msg)
            pltw.datatext.setText(info)
            if not wait(5 + extratim, mainw):
                err = -1
    subw.close()
    return err, msg


def test_PSD(mainw, extratim=0):
    """ Test the 'PSD' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    if mainw.progpath is None:
        # Because it will be impossible to create the temporary file
        return 1, "Cannot create the temporary file"
    path = os.path.join(mainw.testspath, "selftests", "PSD-iso.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Test of the 'PSD' script commands ####\n"
    info += "# Test file: /selftests/PSD-iso.plt\n"
    pltw.datatext.setText(info)
    if not wait(3, mainw):
         err = -1
    else:
         psdnam = "{0}/PSD-PSD.plt".format(mainw.progpath)
         err, msg = PSDcalc(pltw.blklst[1], psdnam, 'D', "halsey")
    if not err:
        info += "\n# Execute 'PSD D halsey' command"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    if not err:
        msg = mainw.loadFile(psdnam)
        if msg == "":
            if not wait(4+extratim, mainw):
                err = -1
            subwpsd = mainw.isAlreadyOpen(psdnam)
            if subwpsd is not None:
                subwpsd.close()
        else:
            err = 1
    subw.close()
    return err, msg


def test_revert(mainw, extratim=0):
    """ Test the 'revert' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "revert.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    mainw.tableTool()
    info = "#### Test of the 'revertv and 'revertb' script commands ####\n"
    info += "# Test file: /selftests/revert.plt\n"
    info += "# Display the data table"
    pltw.datatext.setText(info)
    if not wait(3, mainw):
        err = -1
    else:
        err, msg = revertv(pltw, 'X')
    if not err:
        pltw.updatePlot()
        pltw.tabledlg.table.initTable()
        info += "\n# After executing 'revertv X' command"
        info += "\n# The data in X are now in reverse order"
        pltw.datatext.setText(info)
        if not wait(3, mainw):
            err = -1
    if not err:
        err, msg = revertb(pltw, 0)
    if not err:
        pltw.updatePlot()
        pltw.tabledlg.table.initTable()
        info += "\n\n# After executing 'revertb 1' command"
        info += "\n# The data in X and Y are now in reverse order"
        info += "\n\n# End of the test"
        pltw.datatext.setText(info)
        if not wait(3 + extratim, mainw):
            err = -1
        pltw.tabledlg.close()
    subw.close()
    return err, msg


def test_shift(mainw, extratim=0):
    """ Test the 'shift' script command.

        This also test Cursor and Marker
    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "shift.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Test of the 'shift' script commands ####\n"
    info += "# Test file: /selftests/shift.plt\n"
    pltw.datatext.setText(info)
    if not wait(3, mainw):
        err = -1
    else:
        err = 0
    pltw.dcursor = dCursor(pltw)
    if pltw.dcursor.move(indx=20):
        info += "# Set Cursor at X = 2\n"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    else:
        err = 1
        msg = "Data cursor error"
    if not err:
        mainw.addMark()
        info += "# Set a marker at cursor position\n"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    if not err:
        if pltw.dcursor.move(indx=60):
            info += "# Set Cursor at X = 6\n"
            pltw.datatext.setText(info)
            if not wait(2, mainw):
                err = -1
        else:
            err = 1
            msg = "Data cursor error"
    if not err:
        mainw.addMark()
        info += "# Set a marker at cursor position\n"
        pltw.datatext.setText(info)
        if not wait(2, mainw):
            err = -1
    if not err:
        err, msg = shift(pltw, pltw.curvelist[0], -0.6, 20, 60)
        if not err:
            pltw.updatePlot()
            info += "# Execute 'shift Y -0.6' command"
            info += "\n# End of the test"
            pltw.datatext.setText(info)
            if not wait(3 + extratim, mainw):
                err = -1
    subw.close()
    return err, msg


def test_shrink(mainw, extratim=0):
    """ Test the 'shrink' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "shrink.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Test of the 'shrink' script commands ####\n"
    info += "# Test file: /selftests/shrink.plt\n"
    pltw.datatext.setText(info)
    if not wait(3, mainw):
        err = -1
    else:
        info += "# Execute the command 'shrink TG 5'\n"
        err, msg = shrink(pltw, 'TG', 5)
    if not err:
        info += "\n# {0}\n\n# End of the test".format(msg)
        pltw.datatext.setText(info)
        if not wait(4 + extratim, mainw):
            err = -1
    subw.close()
    return err, msg


def test_sort(mainw, extratim=0):
    """ Test the 'sort' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "sort.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Test of the 'sort' script commands ####\n"
    info += "# Test file: /selftests/sort.plt\n"
    info += "# This is the curve of Y = X*X but X values are not sorted."
    pltw.datatext.setText(info)
    if not wait(4, mainw):
        err = -1
    else:
        err, msg = sort(pltw, 'X')
    if not err:
        pltw.updatePlot()
        info += "\n# After executing 'sort X' command"
        info += "\n\n# End of the test"
        pltw.datatext.setText(info)
        if not wait(4 + extratim, mainw):
            err = -1
    subw.close()
    return err, msg


def test_stats(mainw, extratim=0):
    """ Test the 'stats' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "stats.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Test of the 'stats' script commands ####\n"
    info += "# Test file: /selftests/stats.plt\n"
    info += "# About to execute 'stats Y' command"
    pltw.datatext.setText(info)
    if not wait(3, mainw):
        err = -1
    else:
        err, msg = stats(pltw, 'Y')
    if not err:
        pltw.datatext.setText(msg)
        if not wait(5 + extratim, mainw):
            err = -1
    subw.close()
    return err, msg


def test_swapv(mainw, extratim=0):
    """ Test the 'swapv' script command.

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "swapv.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Test of the 'swapv' script commands ####\n"
    info += "# Test file: /selftests/swapv.plt"
    pltw.datatext.setText(info)
    if not wait(3, mainw):
        err = -1
    else:
        err, ms = swapv(pltw, 'X', 'Y')
    if not err:
        pltw.updatePlot()
        info += "\n# After executing 'swapv X Y' command"
        info += "\n\n# End of the test"
        pltw.datatext.setText(info)
        if not wait(3 + extratim, mainw):
            err = -1
        subw.close()
    return err, msg


# ------------------------------------------------------------------------

sctstnam = ["area", "BET", "clipup & clipdn", "clipx", "delb", "deldupx",
            "delv", "despike", "fft", "IR", "line", "linefit", "lineq",
            "mergeb", "ndec", "newv", "onset", "PSD", "revert", "shift",
            "shrink", "sort", "stats", "swapv"]

sctstfunc = [test_area, test_BET, test_clipupdn, test_clipx, test_delb,
            test_deldupx, test_delv, test_despike, test_fft, test_IR,
            test_line, test_linefit, test_lineq, test_mergeb, test_ndec,
            test_newv, test_onset, test_PSD, test_revert, test_shift,
            test_shrink, test_sort, test_stats, test_swapv ]

# ------------------------------------------------------------------------

def test_sep(mainw, extratim=0):
    """ Test reading files with various separators

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    seplst = ["colon", "comma", "semicolon", "space", "tab"]
    info = "#### Testing various column separators #### \n\n"
    err = 0
    for sep in seplst:
        if sep == "comma":
            ext = "csv"
        else:
            ext = "txt"

        filnam = "sep-{0}.{1}".format(sep, ext)
        path = os.path.join(mainw.testspath, "selftests", filnam)
        msg = mainw.loadFile(path)
        if msg:
            err = 1
        else:
            subw = mainw.getCurrentSubWindow()
            if subw is None:
                return 1, "QtGui.QMdiArea error"
            pltw = subw.widget()
            info += "Reading a file where column separator is {0}\n".format(sep)
            info += "# Test file: /selftests/sep-{0}.txt\n\n".format(sep)
            if sep == "tab":
                info += "# End of the test"
            pltw.datatext.setText(info)
            pltw.datatext.moveCursor(QtGui.QTextCursor.End)
            if not wait(4, mainw):
                err = -1
            subw.close()
        if err:
            break
    return err, msg


def test_JCAMP(mainw, extratim=0):
    """ Test reading JCAMP-DX files

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    err = 0
    chemlst = ["1-Propanol-IR", "EtOH-IR", "acetone-MS", "AcetylAcetone-MS",
               "Ethanol-NMR"]
    for i, chem in enumerate(chemlst):
        filnam = "{0}.jdx".format(chem)
        path = os.path.join(mainw.testspath, "selftests", filnam)
        msg = mainw.loadFile(path)
        if msg != "":
            err = 1
        else:
            subw = mainw.getCurrentSubWindow()
            if subw is None:
                msg = "QtGui.QMdiArea error"
            else:
                pltw = subw.widget()
                info = "#### Reading a JCAMP-DX file ####\n"
                info += "# Test file: /selftests/{0}.jdx\n\n".format(chem)
                if i == len(chemlst) - 1:
                    info += "\n# End of the test"
                pltw.datatext.setText(info)
                if not wait(4, mainw):
                    err = -1
                subw.close()
            if err:
                break
    return err, msg


def test_RRUFF(mainw, extratim=0):
    """ Test reading RRUFF files

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    err = 0
    chemlst = ["TiO2-Anatase-Raman", "TiO2-Rutile-Raman"]
    for i, chem in enumerate(chemlst):
        filnam = "{0}.txt".format(chem)
        path = os.path.join(mainw.testspath, "selftests", filnam)
        msg = mainw.loadFile(path)
        if msg != "":
            err = 1
        else:
            subw = mainw.getCurrentSubWindow()
            if subw is None:
                msg = "QtGui.QMdiArea error"
            else:
                pltw = subw.widget()
                info = "#### Reading a RRUFF Raman file ####\n"
                info += "# Test file: /selftests/{0}.txt\n\n".format(chem)
                if i == len(chemlst) - 1:
                    info += "\n# End of the test"
                pltw.datatext.setText(info)
                if not wait(4, mainw):
                    err = -1
                subw.close()
            if err:
                break
    return err, msg


def test_append(mainw, extratim=0):
    """ Test the Append option in file menu

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "append-1.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    err = 0
    pltw = subw.widget()
    info = "#### Testing Append option in file menu ####\n"
    info += "# Test file: /selftests/append-1.plt\n\n"
    pltw.datatext.setText(info)
    if not wait(3, mainw):
        err = -1
    if not err:
        path = os.path.join(mainw.testspath, "selftests", "append-2.plt")
        msg = mainw.appendFile(path)
        if msg is not None:
            err = 1
        else:
            info += "# The file data in: /selftests/append-2.plt\n"
            info += "# has been added at the end of the initial data\n"
            info += "\n# End of the test"
            pltw.datatext.setText(info)
            if not wait(3, mainw):
                err = -1
    pltw.dirty = False
    subw.close()
    return err, msg


def test_export_csv(mainw, extratim=0):
    """ Test the 'Export to CSV' option in file menu

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "ExportToCSV.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    err = 1
    pltw = subw.widget()
    info = "#### Testing 'Export to CSV' option in file menu ####\n"
    info += "# Test file: /selftests/ExportToCSV.plt\n\n"
    path = "{0}/selftests/exported.csv".format(mainw.testspath)
    if pltw.exportToCSV(0, path):
        info += "The data block has been exported in 'exported.csv'"
        # Read the exported file to check it is ok.
        msg, ftext = pltw.fileToString(path)
        if ftext is not None:
            txtlines = ftext.splitlines()
            msg, blocks, headers = textToData(txtlines, ',')
            if msg == "":
                err = 0
                info += "\n# End of the test"
                pltw.datatext.setText(info)
                if not wait(4, mainw):
                    err = -1
    subw.close()
    return err, msg


def test_export_xls(mainw, extratim=0):
    """ Test the 'Export to Excel' option in file menu

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "2-Block.plt")
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Testing 'Export to Excel' option in file menu ####\n"
    info += "# Test file: /selftests/2-Block.plt\n\n"
    path = "{0}/selftests/2-Sheet.xls".format(mainw.testspath)
    err, msg = mainw.exportExcel(pltw, path)
    if not err:
        info += "Two data blocks have been exported in '2-Sheet.xls'"
        info += "\n# End of the test"
        pltw.datatext.setText(info)
        if not wait(4, mainw):
            err = -1
    subw.close()
    return err, msg


def test_import(mainw, extratim=0):
    """ Test the 'Import from spreadsheet' option in file menu

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", "2-Sheet.xls")
    err, msg = mainw.importxls(path)
    if err:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "#### Testing 'Import from spreadsheet' option in file menu ####\n"
    info += "# 1 - Import from Excel file \n"
    info += "# Test file: /selftests/2-Sheet.xls\n\n"
    if len(pltw.blklst) == 2:
        info += "Two sheets have been imported as two data blocks\n"
        pltw.datatext.setText(info)
        if not wait(4, mainw):
            err = -1
    else:
        err = 1
        msg = "Unexpected data size"
    subw.close()
    if err:
        return err, msg

    path = os.path.join(mainw.testspath, "selftests", "1-Sheet.ods")
    err, msg = mainw.importxls(path)
    if err:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    info = "# 2 - Import from Open Document file (ODF)\n"
    info += "# Test file: /selftests/1-Sheet.ods\n\n"
    if len(pltw.blklst) == 1:
        info += "One sheet has been imported in one data block\n"
        info += "\n# End of the test"
        pltw.datatext.setText(info)
        if not wait(4, mainw):
            err = -1
    else:
        err = 1
        msg = "Unexpected data size"
    subw.close()
    return err, msg


# ------------------------------------------------------------------------

iotstnam = ["test_sep", "test_JCAMP", "test_RRUFF", "test_append",
            "test_export_csv", "test_export_xls", "test_import"]

iotstfunc = [test_sep, test_JCAMP, test_RRUFF, test_append,
             test_export_csv, test_export_xls, test_import]


def convert(mainw, filnam, info, convfunc):
    """ Generic function for testing convert options in convert menu.

    :param mainw: the application MainWindow
    :param filnam: string with the name of the test file
    :param info: string with the informative text to show
    :param convfunc: the conversion function to call
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    path = os.path.join(mainw.testspath, "selftests", filnam)
    msg = mainw.loadFile(path)
    if msg:
        return 1, msg
    subw = mainw.getCurrentSubWindow()
    if subw is None:
        return 1, "QtGui.QMdiArea error"
    pltw = subw.widget()
    text = str(pltw.datatext.toPlainText()).rstrip('\r\n')
    lines = text.split('\n')
    err = 1
    pltw.datatext.setText(info)
    if not wait(2, mainw):
        err = -1
    # remove the file extension
    basename = os.path.splitext(path)[0]
    # the saved file name is obtained by replacement of the extension by .plt
    savename = "{0}.plt".format(basename)
    infnam = os.path.split(path)[1]
    outfnam = os.path.split(savename)[1]
    if convfunc == convertShape:
        msg = convfunc(lines, 2, savename)
    elif convfunc == transpose:
        msg = convfunc(mainw, savename)
    else:
        msg = convfunc(lines, savename)
    subw.close()
    if msg is None:
        # load the converted file
        msg = mainw.loadFile(savename)
        if not msg:
            subw = mainw.isAlreadyOpen(savename)
            if subw is not None:
                err = 0
                info += '{0} has been successfully used to generate {1}\n'.format(infnam, outfnam)
                info += "\n# End of the test"
                pltw = subw.widget()
                pltw.datatext.setText(info)
                if not wait(5, mainw):
                    err = -1
                subw.close()
    return err, msg

# ------------------------------------------------------------------------

def test_CIF(mainw, extratim=0):
    """ Test the 'CIF' option in convert menu

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    filnam = "Si-COD-9008566.cif"
    info = "#### Testing 'CIF' option in convert menu ####\n"
    info += "# This generates a diffraction pattern from a CIF file\n"
    info += "# Loading test file: /selftests/{0}\n\n".format(filnam)
    return convert(mainw, filnam, info, CIF)


def test_DAT(mainw, extratim=0):
    """ Test the 'DAT' option in convert menu

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    filnam = "urea.dat"
    info = "#### Testing 'DAT' option in convert menu ####\n"
    info += "# This converts a '.DAT' diffraction file in a tabular text\n"
    info += "# Loading test file: /selftests/{0}\n\n".format(filnam)
    return convert(mainw, filnam, info, convDAT)


def test_PRF(mainw, extratim=0):
    """ Test the 'PRF' option in convert menu

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    filnam = "Anatase.prf"
    info = "#### Testing 'PRF' option in convert menu ####\n"
    info += "# This converts a FullProf “.PRF” file in a tabular text\n"
    info += "# Loading test file: /selftests/{0}\n\n".format(filnam)
    return convert(mainw, filnam, info, convPRF)


def test_reshape(mainw, extratim=0):
    """ Test the 'Reshape' option in convert menu

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    filnam = "Reshape.txt"
    info = "#### Testing 'Reshape' option in convert menu ####\n"
    info += "# This changes the shape of a data block\n"
    info += "# Loading test file: /selftests/{0}\n\n".format(filnam)
    return convert(mainw, filnam, info, convertShape)


def test_transpose(mainw, extratim=0):
    """ Test the 'Transpose' option in convert menu

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    filnam = "Transpose.txt"
    info = "#### Testing 'Transpose' option in convert menu ####\n"
    info += "# This switches the rows and columns\n"
    info += "# Loading test file: /selftests/{0}\n\n".format(filnam)
    return convert(mainw, filnam, info, transpose)


def test_1Dto2D(mainw, extratim=0):
    """ Test the '1D->2D' option in convert menu

    :param mainw: the application MainWindow
    :return: a tuple (err, msg).
             err = 0 if success, 1 if failure and -1 if interrupted by user.
             msg = contains an error message ("" if not error)
    """
    filnam = "1D.txt"
    info = "#### Testing '1D->2D' option in convert menu ####\n"
    info += "# This adds a X column to a 1D data file\n"
    info += "# Loading test file: /selftests/{0}\n\n".format(filnam)
    return convert(mainw, filnam, info, convert2D)


# ------------------------------------------------------------------------

convtstnam = ["test_CIF", "test_DAT", "test_PRF", "test_reshape",
              "test_transpose", "test_1D->2D"]

convtstfunc = [test_CIF, test_DAT, test_PRF, test_reshape,
              test_transpose, test_1Dto2D]

